"""Evaluation module
This file contains all the methods for evaluating every module in pyjedai.
"""
from decimal import DivisionByZero
from typing import Type
from typing import List, Tuple
from warnings import warn

import matplotlib.pyplot as plt
import networkx as nx
import pandas as pd
import seaborn as sns
import numpy as np

from .datamodel import Data
from .utils import are_matching
from .utils import batch_pairs
from .utils import canonical_swap
from math import inf
from .utils import PredictionData
from .utils import generate_unique_identifier
import random
import matplotlib.pyplot as plt


class Evaluation:
    """Evaluation class. Contains multiple methods for all the fitted & predicted data.
    """
    def __init__(self, data) -> None:
        self.f1: float
        self.recall: float
        self.precision: float
        self.num_of_comparisons: int
        self.total_matching_pairs: int = 0
        self.true_positives: int
        self.true_negatives: int
        self.false_positives: int
        self.false_negatives: int
        self.num_of_true_duplicates: int
        self.data: Data = data

        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to evaluation without a ground-truth file. " + 
                    "Data object has not been initialized with the ground-truth file")

        self.true_positives = self.true_negatives = self.false_positives = self.false_negatives = 0
        self.all_gt_ids = set(self.data._ids_mapping_1.values()) if self.data.is_dirty_er else \
                        set(self.data._ids_mapping_1.values()).union(set(self.data._ids_mapping_2.values()))

    def _set_true_positives(self, true_positives) -> None:
        self.true_positives = true_positives

    def _set_total_matching_pairs(self, total_matching_pairs) -> None:
        self.total_matching_pairs = total_matching_pairs

    def calculate_scores(self, true_positives=None, total_matching_pairs=None) -> None:
        if true_positives is not None:
            self.true_positives = true_positives
            
        if total_matching_pairs is not None:
            self.total_matching_pairs = total_matching_pairs

        if self.total_matching_pairs == 0:
            warn("Evaluation: No matches found", Warning)
            self.num_of_true_duplicates = self.false_negatives \
                = self.false_positives = self.total_matching_pairs \
                    = self.true_positives = self.true_negatives \
                        = self.recall = self.f1 = self.precision = 0
        else:
            self.num_of_true_duplicates = len(self.data.ground_truth)
            self.false_negatives = self.num_of_true_duplicates - self.true_positives
            self.false_positives = self.total_matching_pairs - self.true_positives
            cardinality = (self.data.num_of_entities_1*(self.data.num_of_entities_1-1))/2 \
                if self.data.is_dirty_er else (self.data.num_of_entities_1 * self.data.num_of_entities_2)
            self.true_negatives = cardinality - self.false_negatives - self.num_of_true_duplicates
            self.precision = self.true_positives / self.total_matching_pairs
            self.recall = self.true_positives / self.num_of_true_duplicates
            if self.precision == 0.0 or self.recall == 0.0:
                self.f1 = 0.0
            else:
                self.f1 = 2*((self.precision*self.recall)/(self.precision+self.recall))

    def report(
            self,
            configuration: dict = None,
            export_to_df=False,
            export_to_dict=False,
            with_classification_report=False,
            verbose=True
        ) -> any:

        results_dict = {
                'Precision %': self.precision*100,
                'Recall %': self.recall*100,
                'F1 %': self.f1*100,
                'True Positives': self.true_positives,
                'False Positives': self.false_positives,
                'True Negatives': self.true_negatives,
                'False Negatives': self.false_negatives
            }

        if verbose:
            if configuration:
                print('*' * 123)
                print(' ' * 40, 'Μethod: ', configuration['name'])
                print('*' * 123)
                print(
                    "Method name: " + configuration['name'] +
                    "\nParameters: \n" + ''.join(['\t{0}: {1}\n'.format(k, v) for k, v in configuration['parameters'].items()]) +
                    "Runtime: {:2.4f} seconds".format(configuration['runtime'])
                )
            else:
                print(" " + (configuration['name'] if configuration else "") + " Evaluation \n---")


            print(u'\u2500' * 123)
            print("Performance:\n\tPrecision: {:9.2f}% \n\tRecall:    {:9.2f}%\n\tF1-score:  {:9.2f}%".format(self.precision*100, self.recall*100, self.f1*100))
            print(u'\u2500' * 123)
            if with_classification_report:
                print("Classification report:\n\tTrue positives: {:d}\n\tFalse positives: {:d}\n\tTrue negatives: {:d}\n\tFalse negatives: {:d}\n\tTotal comparisons: {:d}".format(
                    int(self.true_positives), int(self.false_positives), int(self.true_negatives), \
                    int(self.false_negatives), int(self.total_matching_pairs))
                )
                print(u'\u2500' * 123)
                
        if export_to_df:
            pd.set_option("display.precision", 2)
            results = pd.DataFrame.from_dict(results_dict, orient='index').T
            return results

        return results_dict

    def _create_entity_index_from_clusters(
            self,
            clusters: list,
    ) -> dict:
        entity_index = dict()
        for cluster, cluster_id in zip(clusters, range(0, len(clusters))):
            cluster_entities_d1 = 0
            cluster_entities_d2 = 0
            for entity_id in cluster.intersection(self.all_gt_ids):
                entity_index[entity_id] = cluster_id

                if not self.data.is_dirty_er:
                    if entity_id < self.data.dataset_limit:
                        cluster_entities_d1 += 1
                    else:
                        cluster_entities_d2 += 1

            if self.data.is_dirty_er:
                self.total_matching_pairs += len(cluster)*(len(cluster)-1)/2
            else:
                self.total_matching_pairs += cluster_entities_d1*cluster_entities_d2

        return entity_index

    def _create_entity_index_from_blocks(
            self,
            blocks: dict
    ) -> dict:
        entity_index = dict()
        for block_id, block in blocks.items():
            for entity_id in block.entities_D1:
                entity_index.setdefault(entity_id, set())
                entity_index[entity_id].add(block_id)

            if not self.data.is_dirty_er:
                for entity_id in block.entities_D2:
                    entity_index.setdefault(entity_id, set())
                    entity_index[entity_id].add(block_id)

            if self.data.is_dirty_er:
                self.total_matching_pairs += len(block.entities_D1)*(len(block.entities_D1)-1)/2
            else:
                self.total_matching_pairs += len(block.entities_D1)*len(block.entities_D2)

        return entity_index

    def confusion_matrix(self):
        """Generates a confusion matrix based on the classification report.
        """
        heatmap = [
            [int(self.true_positives), int(self.false_positives)],
            [int(self.false_negatives), int(self.true_negatives)]
        ]
        # plt.colorbar(heatmap)
        sns.heatmap(
            heatmap,
            annot=True,
            cmap='Blues',
            xticklabels=['Non-Matching', 'Matching'],
            yticklabels=['Non-Matching', 'Matching'],
            fmt='g'
        )
        plt.title("Confusion Matrix", fontsize=12, fontweight='bold')
        plt.xlabel("Predicted pairs", fontsize=10, fontweight='bold')
        plt.ylabel("Real matching pairs", fontsize=10, fontweight='bold')
        plt.show()
        
    def visualize_roc(self, methods_data : List[dict], proportional : bool =True, drop_tp_indices=True) -> None:
        fig, ax = plt.subplots(figsize=(10, 6))  # set the size of the plot
        colors = []
        normalized_aucs = []
        # for each method layout its plot
        for method_data in methods_data:
            cumulative_recall, normalized_auc = self._generate_auc_data(total_candidates=method_data['total_emissions'], tp_positions=method_data['tp_idx'])
            if(drop_tp_indices):
                del(method_data['tp_idx'])
            method_name=method_data['name']
            method_data['auc'] = normalized_auc
            method_data['recall'] = cumulative_recall[-1] if len(cumulative_recall) != 0 else 0.0
            
            x_values = range(len(cumulative_recall))
            color = "#{:06x}".format(random.randint(0, 0xFFFFFF))
            colors.append(color)
            normalized_aucs.append(normalized_auc)
            if proportional: sizes = [cr * 100 for cr in cumulative_recall]
            else: sizes = [10] * len(cumulative_recall)
            ax.scatter(x_values, cumulative_recall, marker='o', s=0.05, color=color, label=method_name)
            ax.plot(x_values, cumulative_recall, color=color)

        ax.set_xlabel('ec*', fontweight='bold', labelpad=10)
        ax.set_ylabel('Cumulative Recall', fontweight='bold', labelpad=10)
        ax.set_xlim(0, len(cumulative_recall))
        ax.set_ylim(0, 1)

        # add a legend showing the name of each curve and its color
        legend = ax.legend(ncol=2, loc='lower left', title='Methods', bbox_to_anchor=(0, -0.4))
        legend.get_title().set_fontweight('bold')
        legend.get_title().set_fontsize(12)
        plt.setp(legend.get_lines(), linewidth=4)

        # add AUC score legend
        handles, _ = ax.get_legend_handles_labels()
        auc_legend_labels = ['AUC: {:.2f}'.format(nauc) for nauc in normalized_aucs]
        auc_legend = ax.legend(handles, auc_legend_labels, loc='lower left', bbox_to_anchor=(0.5, -0.4), ncol=2, frameon=True, title='AUC', title_fontsize=12)
        auc_legend.get_title().set_fontweight('bold')
        for i, text in enumerate(auc_legend.get_texts()):
            plt.setp(text, color=colors[i])
        ax.add_artist(legend)

        # set the figure background color to the RGB color of the solarized terminal theme
        fig.patch.set_facecolor((0.909, 0.909, 0.909))

        # adjust the margins of the figure to move the graph to the right
        fig.subplots_adjust(left=0.1, bottom=0.3, right=0.9, top=0.9)

        plt.show()

    def calculate_ideal_auc(self, pairs_num : int, true_duplicates_num : int) -> float:
        """Calculates the ideal AUC for the given number of candidate pairs
        Args:
            pairs_num (int): Total number of candidate pairs
            true_duplicates_num (int): The number of true duplicates 
        Returns:
            float: Ideal AUC
        """
        ideal_auc : float

        if(pairs_num == true_duplicates_num):
            ideal_auc = 0.5
        else:
            ideal_auc = (pairs_num % true_duplicates_num) / true_duplicates_num * 0.5
            if(pairs_num > true_duplicates_num): ideal_auc += (pairs_num - true_duplicates_num) / true_duplicates_num

        return ideal_auc

    def _till_full_tps_emission(self) -> bool:
        """Checks if emission should be stopped once all TPs have been found (TPs dict supplied)
        Returns:
            bool: Stop emission on all TPs found / Emit all pairs
        """
        return self._duplicate_emitted is not None
    
    def _all_tps_emitted(self) -> bool:
        """Checks if all TPs have been emitted (Defaults to False in the case of all pairs emission approach)
        Returns:
            bool: All TPs emitted / not emitted
        """
        if(self._till_full_tps_emission()): return self._tps_found >= len(self._duplicate_emitted)
        else: False
        
    def _update_true_positive_entry(self, entity : int, candidate : int) -> None:
        """Updates the checked status of the given true positive

        Args:
            entity (int): Entity ID
            candidate (int): Candidate ID
        """
        if(self._till_full_tps_emission()):
            if(not self._duplicate_emitted[(entity, candidate)]):
                self._duplicate_emitted[(entity, candidate)] = True
                self._tps_found += 1
                return
    

    def calculate_tps_indices(self, pairs : List[Tuple[float, int, int]], duplicate_of : dict = None, duplicate_emitted : dict = None, batch_size : int  = 1) -> Tuple[List[int], int]:
        """
        Args:
            pairs (List[float, int, int]): Candidate pairs to emit in the form [similarity, first dataframe entity ID, second dataframe entity ID]
            duplicate_of (dict, optional): Dictionary of the form [entity ID] -> [IDs of duplicate entities]. Defaults to None.
            duplicate_emitted (dict, optional): Dictionary of the form [true positive pair] -> [emission status: emitted/not]. Defaults to None.
            batch_size (int, optional): Recall update emission rate. Defaults to 1.

        Raises:
            AttributeError: No ground truth has been given
        Returns:
            Tuple[List[int], int]: Indices of true positive duplicates within the candidates list and the total emissions
        """

        if(duplicate_emitted is not None): 
            for pair in duplicate_emitted.keys():
                duplicate_emitted[pair] = False

        if(duplicate_of is None):
            raise AttributeError("Can calculate ROC AUC without a ground-truth file. \
                Data object mush have initialized with the ground-truth file")
        
        self._tps_found : int = 0
        self._duplicate_emitted : dict = duplicate_emitted
        self._tps_indices : List[int] = []
        
        batches = batch_pairs(pairs, batch_size)
        # ideal_auc = self.calculate_ideal_auc(len(pairs), self.num_of_true_duplicates)
        self.total_emissions : int = 0
        for batch in batches:
            for score, entity, candidate in batch:
                if(self._all_tps_emitted()): break                
                if candidate in duplicate_of[entity]:
                    self._update_true_positive_entry(entity, candidate)
                    self._tps_indices.append(self.total_emissions)
                    
            self.total_emissions += 1
            # _normalized_auc += ((_new_recall + _current_recall) / 2) * (_current_batch_size / self.num_of_true_duplicates)
            if(self._all_tps_emitted()): break
            
        # _normalized_auc = 0 if(ideal_auc == 0) else _normalized_auc / ideal_auc
        return self._tps_indices, self.total_emissions
    
    
    def _generate_auc_data(self, total_candidates : int, tp_positions : List[int]) -> Tuple[List[float], float]:
        """Generates the recall axis containing the recall value for each emission and calculates the normalized AUC

        Args:
            total_candidates (int): Total number of pairs emitted
            tp_positions (List[int]): Indices of true positives within the candidate pairs list

        Returns:
            Tuple[List[float], float]: Recall axis and the normalized AUC
        """
        
        _recall_axis : List[float] = []
        _recall : float = 0.0
        _tp_index : int = 0
        _dataset_total_tps : int = len(self.data.ground_truth)
        _total_found_tps : int = len(tp_positions)
                
        for recall_index in range(total_candidates):
            if(_tp_index < _total_found_tps):
                if(recall_index == tp_positions[_tp_index]):
                   _recall =  (_tp_index + 1.0) / _dataset_total_tps
                   _tp_index += 1
            _recall_axis.append(_recall)
            
        _normalized_auc : float = sum(_recall_axis) / (total_candidates + 1.0)
        
        return _recall_axis, _normalized_auc    
    
    
    def visualize_results_roc(self, results : dict, drop_tp_indices=True) -> None:
        """For each of the executed workflows, calculates the cumulative recall and normalized AUC based upon true positive indices.
           Finally, displays the ROC for all of the workflows with proper annotation (each workflow gains a unique identifier).
        Args:
            results (dict): Nested dictionary of the form [dataset] -> [matcher] -> [executed workflows and their info] / [model] -> [executed -//-]
        """
        
        workflows_info : List[Tuple[dict]] = []
        
        for dataset in results:
            matchers = results[dataset]
            for matcher in matchers:
                matcher_info = matchers[matcher]
                if(isinstance(matcher_info, list)):
                    for workflow_info in matcher_info:
                        workflows_info.append((workflow_info))
                else:
                    for model in matcher_info:
                        for workflow_info in matcher_info[model]:
                            workflows_info.append((workflow_info))
                          
        self.visualize_roc(workflows_info, drop_tp_indices=drop_tp_indices)
    
    
    def evaluate_auc_roc(self, matchers : List, batch_size : int = 1, proportional : bool = True, drop_tp_indices=True) -> None:
        """For each matcher, takes its prediction data, calculates cumulative recall and auc, plots the corresponding ROC curve, populates prediction data with performance info
        Args:
            matchers List[ProgressiveMatching]: Progressive Matchers
            batch_size (int, optional): Emitted pairs step at which cumulative recall is recalculated. Defaults to 1.
            proportional (bool) : Proportional Visualization
        Raises:
            AttributeError: No Data object
            AttributeError: No Ground Truth file
        """
        
        if self.data is None:
            raise AttributeError("Can not proceed to AUC ROC evaluation without data object.")

        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to AUC ROC evaluation without a ground-truth file. " +
                    "Data object has not been initialized with the ground-truth file")

        self.matchers_info = []
        
        for matcher in matchers:
            _tp_indices, _total_emissions  = self.calculate_tps_indices(pairs=matcher.pairs, duplicate_of=matcher.duplicate_of, duplicate_emitted=matcher.duplicate_emitted)
            matcher_info = {}
            matcher_info['name'] = generate_unique_identifier()
            matcher_info['total_emissions'] = _total_emissions
            matcher_info['tp_idx'] = _tp_indices
            matcher_info['time'] = matcher.execution_time
            
            matcher_prediction_data : PredictionData = PredictionData(matcher=matcher, matcher_info=matcher_info)
            matcher.set_prediction_data(matcher_prediction_data)
            self.matchers_info.append(matcher_info)

        self.visualize_roc(methods_data=self.matchers_info, drop_tp_indices=drop_tp_indices)



from queue import PriorityQueue
from time import time

import pandas as pd
from networkx import Graph, connected_components
from tqdm.autonotebook import tqdm
from ordered_set import OrderedSet

from .datamodel import Data, PYJEDAIFeature
from .evaluation import Evaluation
from .utils import are_matching


class AbstractClustering(PYJEDAIFeature):
    
    def __init__(self) -> None:
        super().__init__()
        self.data: Data
        self.similarity_threshold: float = 0.1
        
    def evaluate(self,
                 prediction,
                 export_to_df: bool = False,
                 export_to_dict: bool = False,
                 with_classification_report: bool = False,
                 verbose: bool = True) -> any:

        if prediction is None:
            if self.blocks is None:
                raise AttributeError("Can not proceed to evaluation without build_blocks.")
            else:
                eval_blocks = self.blocks
        else:
            eval_blocks = prediction
            
        if self.data is None:
            raise AttributeError("Can not proceed to evaluation without data object.")

        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to evaluation without a ground-truth file. " + 
                    "Data object has not been initialized with the ground-truth file")

        eval_obj = Evaluation(self.data)
        true_positives = 0
        entity_index = eval_obj._create_entity_index_from_clusters(eval_blocks)
        for _, (id1, id2) in self.data.ground_truth.iterrows():
            id1 = self.data._ids_mapping_1[id1]
            id2 = self.data._ids_mapping_1[id2] if self.data.is_dirty_er else self.data._ids_mapping_2[id2]
            if id1 in entity_index and    \
                id2 in entity_index and are_matching(entity_index, id1, id2):
                true_positives += 1
        # print(entity_index)
        eval_obj.calculate_scores(true_positives=true_positives)
        return eval_obj.report(self.method_configuration(),
                                export_to_df,
                                export_to_dict,
                                with_classification_report,
                                verbose)
    
    def stats(self) -> None:
        pass

    def export_to_df(self, prediction: list) -> pd.DataFrame:
        """creates a dataframe for the evaluation report

        Args:
            prediction (any): Predicted clusters

        Returns:
            pd.DataFrame: Dataframe containg evaluation scores and stats
        """
        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to evaluation without a ground-truth file. \
                Data object mush have initialized with the ground-truth file")
        pairs_df = pd.DataFrame(columns=['id1', 'id2'])
        for cluster in prediction:
            lcluster = list(cluster)
            for i1 in range(0, len(lcluster)):
                for i2 in range(i1+1, len(lcluster)):
                    if lcluster[i1] < self.data.dataset_limit:
                        id1 = self.data._gt_to_ids_reversed_1[lcluster[i1]]
                        id2 = self.data._gt_to_ids_reversed_1[lcluster[i2]] if self.data.is_dirty_er else self.data._gt_to_ids_reversed_2[lcluster[i2]]
                    else:
                        id2 = self.data._gt_to_ids_reversed_2[lcluster[i1]]
                        id1 = self.data._gt_to_ids_reversed_1[lcluster[i2]]
                    pairs_df = pd.concat(
                        [pairs_df, pd.DataFrame([{'id1':id1, 'id2':id2}], index=[0])], 
                        ignore_index=True
                    )
        return pairs_df
    
    
    

class ConnectedComponentsClustering(AbstractClustering):
    """Creates the connected components of the graph. \
        Applied to graph created from entity matching. \
        Input graph consists of the entity ids (nodes) and the similarity scores (edges).
    """

    _method_name: str = "Connected Components Clustering"
    _method_short_name: str = "CCC"
    _method_info: str = "Gets equivalence clusters from the " + \
                    "transitive closure of the similarity graph."

    def __init__(self) -> None:
        super().__init__()
        self.similarity_threshold: float

    def process(self, graph: Graph, data: Data, similarity_threshold: float = None) -> list:
        """NetworkX Connected Components Algorithm in the produced graph.

        Args:
            graph (Graph): Consists of the entity ids (nodes) and the similarity scores (edges).

        Returns:
            list: list of clusters
        """
        start_time = time()
        self.data = data
        self.similarity_threshold: float = similarity_threshold
        graph_copy = graph.copy()
        if self.similarity_threshold is not None:
            for x in graph.edges(data=True):
                if x[2]['weight'] < self.similarity_threshold:
                    graph_copy.remove_edge(x[0], x[1])
        clusters = list(connected_components(graph_copy))
        # print(clusters)
        # print("Number of clusters: ", len(clusters))
        resulting_clusters = list(filter(lambda x: len(x) == 2, clusters)) \
                                if not data.is_dirty_er else clusters
        # print("Number of clusters after filtering: ", len(resulting_clusters))
        self.execution_time = time() - start_time
        return resulting_clusters

    def _configuration(self) -> dict:
        return {}

class UniqueMappingClustering(AbstractClustering):
    """Prunes all edges with a weight lower than t, sorts the remaining ones in
        decreasing weight/similarity and iteratively forms a partition for
        the top-weighted pair as long as none of its entities has already
        been matched to some other.
    """

    _method_name: str = "Unique Mapping Clustering"
    _method_short_name: str = "UMC"
    _method_info: str = "Prunes all edges with a weight lower than t, sorts the remaining ones in" + \
                        "decreasing weight/similarity and iteratively forms a partition for" + \
                        "the top-weighted pair as long as none of its entities has already" + \
                        "been matched to some other."

    def __init__(self) -> None:
        """Unique Mapping Clustering Constructor

        Args:
            similarity_threshold (float, optional): Prunes all edges with a weight
                lower than this. Defaults to 0.1.
            data (Data): Dataset module.
        """
        super().__init__()
        self.similarity_threshold: float

    def process(self, graph: Graph, data: Data, similarity_threshold: float = 0.1) -> list:
        """NetworkX Connected Components Algorithm in the produced graph.

        Args:
            graph (Graph): Consists of the entity ids (nodes) and the similarity scores (edges).

        Returns:
            list: list of clusters
        """
        if data.is_dirty_er:
            raise AttributeError("Unique Mapping Clustering can only be performed in Clean-Clean Entity Resolution.")
        self.similarity_threshold: float = similarity_threshold
        
        start_time = time()
        matched_entities = OrderedSet()
        self.data = data
        new_graph = Graph()
        priority_queue = PriorityQueue(maxsize = graph.number_of_edges()*2)
        for x in graph.edges(data=True):
            if x[2]['weight'] > self.similarity_threshold:
                priority_queue.put_nowait((1- x[2]['weight'], x[0], x[1]))

        while not priority_queue.empty():
            sim, entity_1, entity_2 = priority_queue.get()
            if entity_1 in matched_entities or entity_2 in matched_entities:
                continue
            new_graph.add_edge(entity_1, entity_2, weight=sim)
            matched_entities.add(entity_1)
            matched_entities.add(entity_2)

        clusters = ConnectedComponentsClustering().process(new_graph, data, similarity_threshold=None)
        self.execution_time = time() - start_time
        return clusters

    def _configuration(self) -> dict:
        return {}

class ExactClustering(AbstractClustering):
    """Implements an adapted, simplified version of the Exact THRESHOLD algorithm,
        introduced in "Similarity Flooding: A Versatile Graph Matching Algorithm and Its Application to Schema Matching",
        also referred in "BIGMAT: A Distributed Affinity-Preserving Random Walk Strategy for Instance Matching on Knowledge Graphs".
        In essence, it keeps the top-1 candidate per entity, as long as the candidate also considers this node as its top candidate.
    """

    _method_name: str = "Exact Clustering"
    _method_short_name: str = "EC"
    _method_info: str = "Ιmplements an adapted, simplified version of the Exact THRESHOLD algorithm," + \
        "In essence, it keeps the top-1 candidate per entity, as long as the candidate also considers this node as its top candidate."

    def __init__(self, similarity_threshold: float = 0.1) -> None:
        """"""
        super().__init__()
        self.similarity_threshold = similarity_threshold

    def process(self, graph: Graph, data: Data) -> list:
        """
        """
        raise NotImplementedError("Exact Clustering is not implemented yet.")

    def _configuration(self) -> dict:
        return {}



import os
import sys

import pandas as pd
sys.path.append(os.path.join(os.path.dirname(__file__), '../'))

# 
# Datamodel
#
from pyjedai.datamodel import Data

dirty_data = Data(
    dataset_1=pd.read_csv("data/der/cora/cora.csv", sep='|'),
    id_column_name_1='Entity Id',
    ground_truth=pd.read_csv("data/der/cora/cora_gt.csv", sep='|', header=None),
)
dirty_data.print_specs()

def test_datamodel_dirty_er():
    assert dirty_data is not None

clean_clean_data = Data(
    dataset_1=pd.read_csv("data/test/ccer/abt_100.csv", sep='|', engine='python').astype(str),
    id_column_name_1='id',
    dataset_2=pd.read_csv("data/test/ccer/buy_100.csv", sep='|', engine='python').astype(str),
    id_column_name_2='id',
    ground_truth=pd.read_csv("data/test/ccer/gt_100.csv", sep='|', engine='python')
)
clean_clean_data.print_specs()

def test_datamodel_clean_clean_er():
    assert clean_clean_data is not None

def test_standard_blocking():
    from pyjedai.block_building import StandardBlocking
    assert StandardBlocking().build_blocks(dirty_data, tqdm_disable=True) is not None
    assert StandardBlocking().build_blocks(clean_clean_data, tqdm_disable=True) is not None

def test_qgrams_blocking():
    from pyjedai.block_building import QGramsBlocking
    assert QGramsBlocking().build_blocks(dirty_data, tqdm_disable=True) is not None
    assert QGramsBlocking().build_blocks(clean_clean_data, tqdm_disable=True) is not None

def test_extended_qgrams_blocking():
    from pyjedai.block_building import ExtendedQGramsBlocking
    assert ExtendedQGramsBlocking().build_blocks(dirty_data, tqdm_disable=True) is not None
    assert ExtendedQGramsBlocking().build_blocks(clean_clean_data, tqdm_disable=True) is not None
    
def test_suffix_arrays_blocking():
    from pyjedai.block_building import SuffixArraysBlocking
    assert SuffixArraysBlocking().build_blocks(dirty_data, tqdm_disable=True) is not None
    assert SuffixArraysBlocking().build_blocks(clean_clean_data, tqdm_disable=True) is not None
    
def test_extended_suffix_arrays_blocking():
    from pyjedai.block_building import ExtendedSuffixArraysBlocking
    assert ExtendedSuffixArraysBlocking().build_blocks(dirty_data, tqdm_disable=True) is not None
    assert ExtendedSuffixArraysBlocking().build_blocks(clean_clean_data, tqdm_disable=True) is not None



import os
import sys

import pandas as pd
sys.path.append(os.path.join(os.path.dirname(__file__), '../'))

# --------------------------------- #

from pyjedai.datamodel import Data

dirty_data = Data(
    dataset_1=pd.read_csv("data/der/cora/cora.csv", sep='|'),
    id_column_name_1='Entity Id',
    ground_truth=pd.read_csv("data/der/cora/cora_gt.csv", sep='|', header=None),
)
dirty_data.print_specs()

def test_datamodel_dirty_er():
    assert dirty_data is not None

clean_clean_data = Data(
    dataset_1=pd.read_csv("data/test/ccer/abt_100.csv", sep='|', engine='python').astype(str),
    id_column_name_1='id',
    dataset_2=pd.read_csv("data/test/ccer/buy_100.csv", sep='|', engine='python').astype(str),
    id_column_name_2='id',
    ground_truth=pd.read_csv("data/test/ccer/gt_100.csv", sep='|', engine='python')
)
clean_clean_data.print_specs()

def test_datamodel_clean_clean_er():
    assert clean_clean_data is not None

# --------------------------------- #

from pyjedai.block_building import StandardBlocking
dblocks = StandardBlocking().build_blocks(dirty_data)
ccblocks = StandardBlocking().build_blocks(clean_clean_data)

# --------------------------------- #

def test_block_filtering():
    from pyjedai.block_cleaning import BlockFiltering
    assert BlockFiltering().process(dblocks, dirty_data) is not None
    assert BlockFiltering().process(ccblocks, clean_clean_data) is not None

def test_block_purging():
    from pyjedai.block_cleaning import BlockPurging
    assert BlockPurging().process(dblocks, dirty_data) is not None
    assert BlockPurging().process(ccblocks, clean_clean_data) is not None



from abc import ABC, abstractmethod
from itertools import count
from time import time
from typing import Callable, List, Tuple

import matplotlib.pyplot as plt
import os
import json
import optuna
import pandas as pd
from networkx import Graph
from tqdm.autonotebook import tqdm

from .datamodel import Data
from .evaluation import Evaluation
from .block_building import StandardBlocking
from .block_cleaning import BlockFiltering, BlockPurging
from .comparison_cleaning import *
from .matching import EntityMatching
from .clustering import ConnectedComponentsClustering, UniqueMappingClustering
from .vector_based_blocking import EmbeddingsNNBlockBuilding
from .joins import EJoin, TopKJoin

from .prioritization import ProgressiveMatching, BlockIndependentPM, class_references
from .utils import new_dictionary_from_keys, get_class_function_arguments, generate_unique_identifier



plt.style.use('seaborn-whitegrid')

class PYJEDAIWorkFlow(ABC):
    """Main module of the pyjedAI and the simplest way to create an end-to-end ER workflow.
    """

    _id = count()

    def __init__(
            self,
            name: str = None
    ) -> None:
        self.f1: list = []
        self.recall: list = []
        self.precision: list = []
        self.runtime: list = []
        self.configurations: list = []
        self.workflow_exec_time: float
        self._id: int = next(self._id)
        self.name: str = name if name else "Workflow-" + str(self._id)
        self._workflow_bar: tqdm
        self.final_pairs = None

    @abstractmethod
    def run(self,
            data: Data,
            verbose: bool = False,
            with_classification_report: bool = False,
            workflow_step_tqdm_disable: bool = True,
            workflow_tqdm_enable: bool = False) -> None:
        pass

    def _init_experiment(self) -> None:
        self.f1: list = []
        self.recall: list = []
        self.precision: list = []
        self.runtime: list = []
        self.configurations: list = []
        self.workflow_exec_time: float

    def visualize(
            self,
            f1: bool = True,
            recall: bool = True,
            precision: bool = True,
            separate: bool = False
    ) -> None:
        """Performance Visualization of the workflow.

        Args:
            f1 (bool, optional): F-Measure. Defaults to True.
            recall (bool, optional): Recall. Defaults to True.
            precision (bool, optional): Precision. Defaults to True.
            separate (bool, optional): Separate plots. Defaults to False.
        """
        method_names = [conf['name'] for conf in self.configurations]
        exec_time = []
        prev = 0

        for i, _ in enumerate(self.runtime):
            exec_time.append(prev + self.runtime[i])
            prev = exec_time[i]

        if separate:
            fig, axs = plt.subplots(nrows=2, ncols=2, figsize=(10, 8))
            fig.suptitle(self.name + " Visualization", fontweight='bold', fontsize=14)
            fig.subplots_adjust(top=0.88)
            axs[0, 0].plot(method_names,
                           self.precision,
                           linewidth=2.0,
                           label="Precision",
                           marker='o',
                           markersize=10)
            axs[0, 0].set_ylabel("Scores %", fontsize=12)
            axs[0, 0].set_title("Precision", fontsize=12)
            axs[0, 1].plot(method_names,
                           self.recall,
                           linewidth=2.0,
                           label="Recall",
                           marker='*',
                           markersize=10)
            axs[0, 1].set_ylabel("Scores %", fontsize=12)
            axs[0, 1].set_title("Recall", fontsize=12)            
            axs[1, 0].plot(method_names,
                           self.f1,
                           linewidth=2.0,
                           label="F1-Score",
                           marker='x',
                           markersize=10)
            axs[1, 0].set_ylabel("Scores %", fontsize=12)
            axs[1, 0].set_title("F1-Score", fontsize=12)
            # axs[0, 0].legend(loc='lower right')
            axs[1, 1].plot(method_names,
                           exec_time,
                           linewidth=2.0,
                           label="Time",
                           marker='.',
                           markersize=10,
                           color='r')
            axs[1, 1].set_ylabel("Time (sec)", fontsize=12)
            axs[1, 1].set_title("Execution time", fontsize=12)
            fig.autofmt_xdate()
        else:
            fig, axs = plt.subplots(nrows=2, ncols=1, figsize=(10, 8))
            fig.suptitle(self.name + " Visualization", fontweight='bold', fontsize=14)
            fig.subplots_adjust(top=0.88)
            if precision:
                axs[0].plot(method_names,
                            self.precision,
                            linewidth=2.0,
                            label="Precision",
                            marker='o',
                            markersize=10)
            if recall:
                axs[0].plot(method_names,
                            self.recall,
                            linewidth=2.0,
                            label="Recall",
                            marker='*',
                            markersize=10)
            if f1:
                axs[0].plot(method_names,
                            self.f1, linewidth=2.0,
                            label="F1-Score",
                            marker='x',
                            markersize=10)
            axs[0].set_xlabel("Models", fontsize=12)
            axs[0].set_ylabel("Scores %", fontsize=12)
            axs[0].set_title("Performance per step", fontsize=12)
            axs[0].legend(loc='lower right')
            exec_time = []
            prev = 0
            for i, _ in enumerate(self.runtime):
                exec_time.append(prev + self.runtime[i])
                prev = exec_time[i]
            axs[1].plot(method_names,
                        exec_time,
                        linewidth=2.0,
                        label="F1-Score",
                        marker='.',
                        markersize=10,
                        color='r')
            axs[1].set_ylabel("Time (sec)", fontsize=12)
            axs[1].set_title("Execution time", fontsize=12)
            fig.autofmt_xdate()
        plt.show()

    def to_df(self) -> pd.DataFrame:
        """Transform results into a pandas.DataFrame

        Returns:
            pd.DataFrame: Results
        """
        workflow_df = pd.DataFrame(
            columns=['Algorithm', 'F1', 'Recall', 'Precision', 'Runtime (sec)', 'Params'])
        workflow_df['F1'], workflow_df['Recall'], \
        workflow_df['Precision'], workflow_df['Runtime (sec)'] = \
            self.f1, self.recall, self.precision, self.runtime
        workflow_df['Algorithm'] = [c['name'] for c in self.configurations]
        workflow_df['Params'] = [c['parameters'] for c in self.configurations]

        return workflow_df

    def export_pairs(self) -> pd.DataFrame:
        """Export pairs to file.

        Returns:
            pd.DataFrame: pairs as a DataFrame
        """
        return write(self.final_pairs, self.data)

    def _save_step(self, results: dict, configuration: dict) -> None:
        self.f1.append(results['F1 %'])
        self.recall.append(results['Recall %'])
        self.precision.append(results['Precision %'])
        self.configurations.append(configuration)
        self.runtime.append(configuration['runtime'])

    def get_final_scores(self) -> Tuple[float, float, float]:
        """Final scores in the last step of the workflow.

        Returns:
            Tuple[float, float, float]: F-Measure, Precision, Recall.
        """
        return self.f1[-1], self.precision[-1], self.recall[-1]
class ProgressiveWorkFlow(PYJEDAIWorkFlow):
    """Main module of the pyjedAI and the simplest way to create an end-to-end PER workflow.
    """

    def __init__(
            self,
            name: str = None
    ) -> None:
        self.f1: list = []
        self.recall: list = []
        self.precision: list = []
        self.runtime: list = []
        self.configurations: list = []
        self.workflow_exec_time: float
        self._id: int = next(self._id)
        self.name: str = name if name else "Workflow-" + str(self._id)
        self._workflow_bar: tqdm
        self.final_pairs = None

    def run(self,
            data: Data,
            verbose: bool = False,
            with_classification_report: bool = False,
            workflow_step_tqdm_disable: bool = True,
            workflow_tqdm_enable: bool = False,
            block_building : dict = None,
            block_purging : dict = None,
            block_filtering : dict = None,
            **matcher_arguments
        ) -> None:
        """Main function for creating an Progressive ER workflow.

        Args:
            data (Data): Dataset Module, used to derive schema-awereness status
            verbose (bool, optional): Print detailed report for each step. Defaults to False.
            with_classification_report (bool, optional): Print pairs counts. Defaults to False.
            workflow_step_tqdm_disable (bool, optional):  Tqdm progress bar in each step. Defaults to True.
            workflow_tqdm_enable (bool, optional): Overall progress bar. Defaults to False.
            number_of_nearest_neighbors (int, optional): Number of nearest neighbours in cardinality based algorithms. Defaults to None.
            indexing (str, optional): Inorder/Reverse/Bilateral indexing of datasets. Defaults to None.
            similarity_function (str, optional): Function used to evaluate the similarity of two vector based representations of entities. Defaults to None.
            language_model (str, optional): Language model used to vectorize the entities. Defaults to None.
            tokenizer (str, optional): Text tokenizer used. Defaults to None.
            weighting_scheme (str, optional): Scheme used to evaluate the weight between nodes of intermediate representation graph. Defaults to None.
            block_building (dict, optional): Algorithm and its parameters used to construct the blocks. Defaults to None.
            block_purging (dict, optional): Algorithm and its parameters used to delete obsolete blocks. Defaults to None.
            block_filtering (dict, optional): Algorithm and its parameters used to lower the cardinality of blocks. Defaults to None.
            window_size (dict, optional): Window size in the Sorted Neighborhood Progressive ER workflows. Defaults to None.
        """
        self.block_building, self.block_purging, self.block_filtering, self.algorithm = \
        block_building, block_purging, block_filtering, matcher_arguments['algorithm']
        steps = [self.block_building, self.block_purging, self.block_filtering, self.algorithm]
        num_of_steps = sum(x is not None for x in steps)
        self._workflow_bar = tqdm(total=num_of_steps,
                                  desc=self.name,
                                  disable=not workflow_tqdm_enable)
         
        self.data : Data = data
        self._init_experiment()
        start_time = time()
        self.matcher_arguments = matcher_arguments
        self.matcher_name = self.matcher_arguments['matcher']
        self.dataset_name = self.matcher_arguments['dataset']
        matcher = class_references[matcher_arguments['matcher']]
        self.constructor_arguments = new_dictionary_from_keys(dictionary=self.matcher_arguments, keys=get_class_function_arguments(class_reference=matcher, function_name='__init__'))
        self.predictor_arguments = new_dictionary_from_keys(dictionary=self.matcher_arguments, keys=get_class_function_arguments(class_reference=matcher, function_name='predict'))
        print(self.constructor_arguments)
        print(self.predictor_arguments)
        
        progressive_matcher : ProgressiveMatching = matcher(**self.constructor_arguments)
        self.progressive_matcher : ProgressiveMatching = progressive_matcher
        #
        # Block Building step: Only one algorithm can be performed
        #
        block_building_method = (self.block_building['method'](**self.block_building["params"]) \
                                                    if "params" in self.block_building \
                                                    else self.block_building['method']()) if self.block_building \
                                                    else (None if not self._blocks_required() else StandardBlocking())

        bblocks = None
        block_building_blocks = None
        if block_building_method:
            block_building_blocks = \
                block_building_method.build_blocks(data,
                                                attributes_1=self.block_building["attributes_1"] \
                                                                    if(self.block_building is not None and "attributes_1" in self.block_building) else None,
                                                    attributes_2=self.block_building["attributes_2"] \
                                                                    if(self.block_building is not None and "attributes_2" in self.block_building) else None,
                                                    tqdm_disable=workflow_step_tqdm_disable)
            self.final_pairs = bblocks = block_building_blocks
            res = block_building_method.evaluate(block_building_blocks,
                                                export_to_dict=True,
                                                with_classification_report=with_classification_report,
                                                verbose=verbose)
            self._save_step(res, block_building_method.method_configuration())
            self._workflow_bar.update(1)

        if(block_building_blocks is not None):
            #
            # Block Purging step [optional]
            #
            bblocks = block_building_blocks
            block_purging_blocks = None
            if(self.block_purging is not None):
                block_purging_method = self.block_purging['method'](**self.block_purging["params"]) \
                                                if "params" in self.block_purging \
                                                else self.block_purging['method']()
                block_purging_blocks = block_purging_method.process(bblocks,
                                                                    data,
                                                                    tqdm_disable=workflow_step_tqdm_disable)
                self.final_pairs = bblocks = block_purging_blocks
                res = block_purging_method.evaluate(bblocks,
                                                    export_to_dict=True,
                                                    with_classification_report=with_classification_report,
                                                    verbose=verbose)
                self._save_step(res, block_purging_method.method_configuration())
                self._workflow_bar.update(1)
            #
            # Block Filtering step [optional]
            #
            block_filtering_blocks = None
            if(self.block_filtering is not None):
                block_filtering_method = self.block_filtering['method'](**self.block_filtering["params"]) \
                                                if "params" in self.block_filtering \
                                                else self.block_filtering['method']()
                block_filtering_blocks = block_filtering_method.process(bblocks,
                                                                        data,
                                                                        tqdm_disable=workflow_step_tqdm_disable)
                self.final_pairs = bblocks = block_filtering_blocks
                res = block_filtering_method.evaluate(bblocks,
                                                    export_to_dict=True,
                                                    with_classification_report=with_classification_report,
                                                    verbose=verbose)
                self._save_step(res, block_filtering_method.method_configuration())
                self._workflow_bar.update(1)

        #
        # Progressive Matching step
        #
        self.final_pairs : List[Tuple[float, int, int]] = progressive_matcher.predict(data=data, blocks=bblocks, **self.predictor_arguments)
        evaluator = Evaluation(self.data)
        self.tp_indices, self.total_emissions = evaluator.calculate_tps_indices(pairs=self.final_pairs,duplicate_of=progressive_matcher.duplicate_of, duplicate_emitted=progressive_matcher.duplicate_emitted)
        self.total_candidates = len(self.final_pairs)       
        self._workflow_bar.update(1)
        self.workflow_exec_time = time() - start_time

    def _blocks_required(self):
        return not isinstance(self.progressive_matcher, BlockIndependentPM)

    def _init_experiment(self) -> None:
        self.f1: list = []
        self.recall: list = []
        self.precision: list = []
        self.runtime: list = []
        self.configurations: list = []
        self.workflow_exec_time: float

    def visualize(
            self,
            f1: bool = True,
            recall: bool = True,
            precision: bool = True,
            separate: bool = False
    ) -> None:
        pass

    def to_df(self) -> pd.DataFrame:
        pass

    def export_pairs(self) -> pd.DataFrame:
        pass

    def _save_step(self, results: dict, configuration: dict) -> None:
        pass

    def get_final_scores(self) -> Tuple[float, float, float]:
        pass
    
    def retrieve_matcher_workflows(self, workflows : dict, arguments : dict) -> list:
        """Retrieves the list of already executed workflows for the matcher/model of current workflow 

        Args:
            workflows (dict): Dictionary of script's executed workflows' information
            arguments (dict): Arguments that have been supplied for current workflow execution

        Returns:
            list: List of already executed workflows for given workflow's arguments' matcher/model
        """
        dataset : str = self.dataset_name
        matcher : str = self.matcher_name
        
        workflows[dataset] = workflows[dataset] if dataset in workflows else dict()
        matcher_results = workflows[dataset]
        matcher_results[matcher] = matcher_results[matcher] if matcher in matcher_results \
                                else ([] if('language_model' not in arguments) else {})
                
        matcher_info = matcher_results[matcher]
        workflows_info = matcher_info
        if(isinstance(matcher_info, dict)):
            lm_name = arguments['language_model']
            matcher_info[lm_name] = matcher_info[lm_name] if lm_name in matcher_info else []
            workflows_info = matcher_info[lm_name]  
            
        return workflows_info
    
    
    
    def save(self, arguments : dict, path : str = None, results = None) -> dict:
        """Stores argument / execution information for current workflow within a workflows dictionary.
        
        Args:
            arguments (dict): Arguments that have been supplied for current workflow execution
            path (str): Path where the workflows results are stored at (Default to None),
            results (str): A dictionary of workflows results at which we want to store current workflow's arguments/info
        Returns:
            dict: Dictionary containing the information about the given workflow
        """
        if(path is None and results is None):
            raise ValueError(f"No dictionary path or workflows dictionary given - Cannot save workflow.")
        
        if(results is not None):
            workflows = results
        elif(not os.path.exists(path) or os.path.getsize(path) == 0):
            workflows = {}
        else:
            with open(path, 'r', encoding="utf-8") as file:
                workflows = json.load(file)
                
        category_workflows = self.retrieve_matcher_workflows(workflows=workflows, arguments=arguments)
        self.save_workflow_info(arguments=arguments) 
        category_workflows.append(self.info)
        
        if(path is not None):
            with open(path, 'w', encoding="utf-8") as file:
                json.dump(workflows, file, indent=4)
            
        return self.info
    
    def save_workflow_info(self, arguments : dict) -> dict:
        """Stores current workflow argument values and execution related data (like execution time and total emissions)

        Args:
            arguments (dict): Arguments that were passed to progressive workflow at hand
        """
        
        workflow_info : dict = {k: v for k, v in arguments.items()}
        workflow_info['total_candidates'] = self.total_candidates
        workflow_info['total_emissions'] = self.total_emissions
        workflow_info['time'] = self.workflow_exec_time
        workflow_info['name'] = generate_unique_identifier()
        workflow_info['tp_idx'] = self.tp_indices
        workflow_info['dataset'] = self.dataset_name
        workflow_info['matcher'] = self.matcher_name

        self.info = workflow_info  
    
    def print_info(self, info : dict):
        for attribute in info:
            value = info[attribute]
            if(attribute != 'tp_idx'):
                print(f"{attribute} : {value}")
            else:
                print(f"true_positives : {len(value)}")
    

def compare_workflows(workflows: List[PYJEDAIWorkFlow], with_visualization=True) -> pd.DataFrame:
    """Compares workflows by creating multiple plots and tables with results.

    Args:
        workflows (List[PYJEDAIWorkFlow]): Different workflows
        with_visualization (bool, optional): Diagram generation. Defaults to True.

    Returns:
        pd.DataFrame: Results
    """
    workflow_df = pd.DataFrame(columns=['Name', 'F1', 'Recall', 'Precision', 'Runtime (sec)'])
    if with_visualization:
        fig, axs = plt.subplots(nrows=2, ncols=2, figsize=(10, 8))
        fig.suptitle("Workflows Performance Visualization", fontweight='bold', fontsize=14)
        fig.subplots_adjust(top=0.88)
        axs[0, 0].set_ylabel("Scores %", fontsize=12)
        axs[0, 0].set_title("Precision", fontsize=12)
        axs[0, 0].set_ylim([0, 100])
        axs[0, 1].set_ylabel("Scores %", fontsize=12)
        axs[0, 1].set_title("Recall", fontsize=12)
        axs[0, 1].set_ylim([0, 100])
        axs[1, 0].set_ylabel("Scores %", fontsize=12)
        axs[1, 0].set_title("F1-Score", fontsize=12)
        axs[1, 0].set_ylim([0, 100])
        axs[1, 1].set_ylabel("Time (sec)", fontsize=12)
        axs[1, 1].set_title("Execution time", fontsize=12)

    for w in workflows:
        workflow_df.loc[len(workflow_df)] = \
            [w.name, w.f1[-1], w.recall[-1], w.precision[-1], w.workflow_exec_time]

    if with_visualization:
        axs[0, 0].bar(workflow_df['Name'],
                      workflow_df['Precision'],
                      label=workflow_df['Name'],
                      color='b')
        axs[0, 1].bar(workflow_df['Name'],
                      workflow_df['Recall'],
                      label=workflow_df['Name'],
                      color='g')
        axs[1, 0].bar(workflow_df['Name'], workflow_df['F1'], color='orange')
        axs[1, 1].bar(workflow_df['Name'], workflow_df['Runtime (sec)'], color='r')
    fig.autofmt_xdate()
    plt.show()

    return workflow_df

class OptimizeWorkflow:
    """Optuna Framework for GridSearch/RandomSearch/Prunning in a given pyjedai workflow.
    """

    _id = count()

    def __init__(
            self,
            block_building: Callable, # Mandatory: one method
            entity_matching: Callable, # Mandatory: one method
            block_cleaning: List[Callable] = None, # Optional: multiple methods
            comparison_cleaning: List[Callable] = None, # Optional: multiple methods
            clustering: Callable = None, # Optional: One method
            joins: Callable = None, # Optional: One method
            name: str = None
    ) -> None:
        self.block_cleaning, self.block_building, self.comparison_cleaning, \
            self.clustering, self.joins, self.entity_matching = \
            block_cleaning, block_building, comparison_cleaning, clustering, joins, entity_matching
        self.f1: float
        self.recall: float
        self.precision: float
        self.runtime: float
        self.configurations: float
        self.workflow_exec_time: float
        self._id: int = next(self._id)
        self.name: str = name if name else "OptWorkflow-" + str(self._id)
        self._workflow_bar: tqdm

    def objective(self, target_score: str = 'f1') -> any:
        """Optuna objective function

            Returns:
                 One or more from F1,Recall,Precision
        """
        pass
        # if target_score  == 'f1':
        #     return f1
        # elif target_score  == 'recall':
        #     return recall
        # else:
        #     return precision

    def run(
            self,
            data: Data,
            num_of_trials = 30,
            pruner: optuna.pruners = optuna.pruners.NopPruner,
            sampler: optuna.samplers = optuna.samplers.BaseSampler,
            study_name = "pyjedai_study",
            storage_name = "pyjedai_storage_trials",
            target_score = "f1",
            load_if_exists=True,
            verbose=False,
            tqdm_disable=False,
            workflow_tqdm_enable=False,
            optuna_tqdm_enable=True
    ) -> pd.DataFrame:
        """Executes the experiments based on a workflow

        Args:
            data (Data): _description_
            num_of_trials (int, optional): _description_. Defaults to 30.
            pruner (optuna.pruners, optional): _description_. Defaults to optuna.pruners.NopPruner.
            sampler (optuna.samplers, optional): _description_. Defaults to optuna.samplers.BaseSampler.
            study_name (str, optional): _description_. Defaults to "pyjedai_study".
            storage_name (str, optional): _description_. Defaults to "pyjedai_storage_trials".
            target_score (str, optional): _description_. Defaults to "f1".
            load_if_exists (bool, optional): _description_. Defaults to True.
            verbose (bool, optional): _description_. Defaults to False.
            tqdm_disable (bool, optional): _description_. Defaults to False.
            workflow_tqdm_enable (bool, optional): _description_. Defaults to False.
            optuna_tqdm_enable (bool, optional): _description_. Defaults to True.

        Returns:
            pd.DataFrame: _description_
        """
        study = optuna.create_study(
            directions=["maximize"],
            study_name=study_name,
            storage=storage_name,
            load_if_exists=load_if_exists
        )
        print("Optuna trials starting")
        study.optimize(
            self.objective, 
            n_trials=num_of_trials, 
            show_progress_bar=optuna_tqdm_enable
        )
        print("Optuna trials finished")

    def visualize(self, with_plotly=True):
        pass
    
    def get_best_trial(self):
        pass
    
    def to_df():
        pass

class BlockingBasedWorkFlow(PYJEDAIWorkFlow):
    """Blocking-based workflow.
    """

    def __init__(
            self,
            block_building: dict = None,
            entity_matching: dict = None,
            block_cleaning: dict = None,
            comparison_cleaning: dict = None,
            clustering: dict = None,
            joins: dict = None,
            name: str = None
    ) -> None:
        super().__init__()
        self.block_cleaning, self.block_building, self.comparison_cleaning, \
            self.clustering, self.joins, self.entity_matching = \
            block_cleaning, block_building, comparison_cleaning, clustering, joins, entity_matching
        self.name: str = name if name else "BlockingBasedWorkFlow-" + str(self._id)

    def run(self,
            data: Data,
            verbose: bool = False,
            with_classification_report: bool = False,
            workflow_step_tqdm_disable: bool = True,
            workflow_tqdm_enable: bool = False
        ) -> None:
        """Main function for creating an Entity resolution workflow.

        Args:
            data (Data): Dataset module.
            verbose (bool, optional): Print detailed report for each step. Defaults to False.
            with_classification_report (bool, optional): Print pairs counts. Defaults to False.
            workflow_step_tqdm_disable (bool, optional):  Tqdm progress bar in each step. Defaults to True.
            workflow_tqdm_enable (bool, optional): Overall progress bar. Defaults to False.
        """
        steps = [self.block_building, self.entity_matching, self.clustering, self.joins, self.block_cleaning, self.comparison_cleaning]
        num_of_steps = sum(x is not None for x in steps)
        self._workflow_bar = tqdm(total=num_of_steps,
                                  desc=self.name,
                                  disable=not workflow_tqdm_enable)
        self.data = data
        self._init_experiment()
        start_time = time()
        #
        # Block building step: Only one algorithm can be performed
        #
        block_building_method = self.block_building['method'](**self.block_building["params"]) \
                                                    if "params" in self.block_building \
                                                    else self.block_building['method']()

        block_building_blocks = \
            block_building_method.build_blocks(data,
                                               attributes_1=self.block_building["attributes_1"] \
                                                                if "attributes_1" in self.block_building else None,
                                                attributes_2=self.block_building["attributes_2"] \
                                                                if "attributes_2" in self.block_building else None,
                                                tqdm_disable=workflow_step_tqdm_disable)
        self.final_pairs = block_building_blocks
        res = block_building_method.evaluate(block_building_blocks,
                                            export_to_dict=True,
                                            with_classification_report=with_classification_report,
                                            verbose=verbose)
        self._save_step(res, block_building_method.method_configuration())
        self._workflow_bar.update(1)
        #
        # Block cleaning step [optional]: Multiple algorithms
        #
        block_cleaning_blocks = None
        if self.block_cleaning:
            if isinstance(self.block_cleaning, dict):
                self.block_cleaning = list(self.block_cleaning)
            bblocks = block_building_blocks
            for block_cleaning in self.block_cleaning:
                block_cleaning_method = block_cleaning['method'](**block_cleaning["params"]) \
                                                    if "params" in block_cleaning \
                                                    else block_cleaning['method']()
                block_cleaning_blocks = block_cleaning_method.process(bblocks,
                                                                      data,
                                                                      tqdm_disable=workflow_step_tqdm_disable)
                
                self.final_pairs = bblocks = block_cleaning_blocks
                res = block_cleaning_method.evaluate(bblocks,
                                                    export_to_dict=True,
                                                    with_classification_report=with_classification_report,
                                                    verbose=verbose)
                self._save_step(res, block_cleaning_method.method_configuration())
                self._workflow_bar.update(1)
        #
        # Comparison cleaning step [optional]
        #
        comparison_cleaning_blocks = None
        if self.comparison_cleaning:
            comparison_cleaning_method = self.comparison_cleaning['method'](**self.comparison_cleaning["params"]) \
                                            if "params" in self.comparison_cleaning \
                                            else self.comparison_cleaning['method']()
            self.final_pairs = \
            comparison_cleaning_blocks = \
            comparison_cleaning_method.process(block_cleaning_blocks if block_cleaning_blocks is not None \
                                                    else block_building_blocks,
                                                data,
                                                tqdm_disable=workflow_step_tqdm_disable)
            res = comparison_cleaning_method.evaluate(comparison_cleaning_blocks,
                                                      export_to_dict=True,
                                                      with_classification_report=with_classification_report,
                                                      verbose=verbose)
            self._save_step(res, comparison_cleaning_method.method_configuration())
            self._workflow_bar.update(1)
        #
        # Entity Matching step
        #
        entity_matching_method = self.entity_matching['method'](**self.entity_matching["params"]) \
                                        if "params" in self.entity_matching \
                                        else self.entity_matching['method']()
                                        
        if "exec_params" not in self.entity_matching:
            self.final_pairs = em_graph = entity_matching_method.predict(
                comparison_cleaning_blocks if comparison_cleaning_blocks is not None \
                    else block_building_blocks,
                data,
                tqdm_disable=workflow_step_tqdm_disable)
        else:
            self.final_pairs = em_graph = entity_matching_method.predict(
                comparison_cleaning_blocks if comparison_cleaning_blocks is not None \
                    else block_building_blocks,
                data,
                tqdm_disable=workflow_step_tqdm_disable,
                **self.entity_matching["exec_params"])

        res = entity_matching_method.evaluate(em_graph,
                                                export_to_dict=True,
                                                with_classification_report=with_classification_report,
                                                verbose=verbose)
        self._save_step(res, entity_matching_method.method_configuration())
        self._workflow_bar.update(1)
        #
        # Clustering step [optional]
        #
        if self.clustering:
            clustering_method = self.clustering['method'](**self.clustering["params"]) \
                                            if "params" in self.clustering \
                                            else self.clustering['method']()
            if "exec_params" not in self.clustering:
                self.final_pairs = components = clustering_method.process(em_graph, data)
            else:
                self.final_pairs = components = clustering_method.process(em_graph, data, **self.clustering["exec_params"])
            
            res = clustering_method.evaluate(components,
                                            export_to_dict=True,
                                            with_classification_report=False,
                                            verbose=verbose)
            self._save_step(res, clustering_method.method_configuration())
            self.workflow_exec_time = time() - start_time
            self._workflow_bar.update(1)
        # self.runtime.append(self.workflow_exec_time)

    ############################################
    #  Pre-defined workflows same as JedAI     #
    ############################################

    def best_blocking_workflow_ccer(self) -> None:
        """Best CC-ER workflow.

        Returns:
            PYJEDAIWorkFlow: Best workflow
        """
        self.block_building = dict(method=StandardBlocking)
        self.block_cleaning = [dict(
            method=BlockFiltering,
            params=dict(ratio=0.9)
        )]
        self.comparison_cleaning = dict(method=WeightedEdgePruning, params=dict(weighting_scheme='EJS'))
        self.entity_matching = dict(method=EntityMatching,
                                    params=dict(metric='cosine',
                                                    tokenizer='char_tokenizer', 
                                                    vectorizer='tfidf',
                                                    qgram=3,
                                                    similarity_threshold=0.0))
        self.clustering = dict(method=UniqueMappingClustering, 
                               exec_params=dict(similarity_threshold=0.17))
        self.name="best-ccer-workflow"

    def best_blocking_workflow_der(self) -> None:
        """Best D-ER workflow.

        Returns:
            PYJEDAIWorkFlow: Best workflow
        """
        self.block_building = dict(method=StandardBlocking)
        self.block_cleaning = [
            dict(method=BlockPurging, params=dict(smoothing_factor=1.0)),
            dict(method=BlockFiltering)
        ]
        self.comparison_cleaning = dict(method=CardinalityNodePruning,
                                    params=dict(weighting_scheme='JS'))
        self.entity_matching = dict(method=EntityMatching, 
                                    params=dict(metric='cosine',
                                                similarity_threshold=0.55))
        self.clustering = dict(method=ConnectedComponentsClustering),
        self.name="best-der-workflow"

    def default_blocking_workflow_ccer(self) -> None:
        """Default CC-ER workflow.
        """
        self.block_building = dict(method=StandardBlocking)
        self.block_cleaning = [
                dict(method=BlockPurging, params=dict(smoothing_factor=1.0)),
                dict(method=BlockFiltering)
            ]
        self.comparison_cleaning = dict(method=CardinalityNodePruning)
        self.entity_matching = dict(method=EntityMatching,
                                    metric='cosine',
                                    similarity_threshold=0.55)
        self.clustering = dict(method=UniqueMappingClustering)
        self.name="default-ccer-workflow"

    def default_blocking_workflow_der(self) -> None:
        """Default D-ER workflow.

        Returns:
            PYJEDAIWorkFlow: Best workflow
        """
        self.block_building = dict(method=StandardBlocking)
        self.block_cleaning = [
                dict(method=BlockPurging, params=dict(smoothing_factor=1.0)),
                dict(method=BlockFiltering)
        ]
        self.comparison_cleaning = dict(method=CardinalityNodePruning, 
                                        params=dict(weighting_scheme='JS'))
        self.entity_matching = dict(method=EntityMatching,
                                    params=dict(metric='cosine', 
                                                similarity_threshold=0.55))
        self.name="default-der-workflow"
 
class EmbeddingsNNWorkFlow(PYJEDAIWorkFlow):
    """Blocking-based workflow.
    """

    def __init__(
            self,
            block_building: dict,
            clustering: dict = None,
            name: str = None
    ) -> None:
        super().__init__()
        self.block_building, self.clustering = block_building, clustering
        self.name: str = name if name else "Workflow-" + str(self._id)

    def run(self,
            data: Data,
            verbose: bool = False,
            with_classification_report: bool = False,
            workflow_step_tqdm_disable: bool = False,
            workflow_tqdm_enable: bool = False
        ) -> None:
        """Main function for creating an Entity resolution workflow.

        Args:
            data (Data): Dataset module.
            verbose (bool, optional): Print detailed report for each step. Defaults to False.
            with_classification_report (bool, optional): Print pairs counts. Defaults to False.
            workflow_step_tqdm_disable (bool, optional):  Tqdm progress bar in each step. Defaults to True.
            workflow_tqdm_enable (bool, optional): Overall progress bar. Defaults to False.
        """
        steps = [self.block_building, self.clustering]
        num_of_steps = sum(x is not None for x in steps)
        self._workflow_bar = tqdm(total=num_of_steps,
                                  desc=self.name,
                                  disable=not workflow_tqdm_enable)
        self.data = data
        self._init_experiment()
        start_time = time()
        #
        # Block building step: Only one algorithm can be performed
        #
        block_building_method = self.block_building['method'](**self.block_building["params"]) \
                                                    if "params" in self.block_building \
                                                    else self.block_building['method']()

        
        if "exec_params" not in self.block_building:
            block_building_blocks, em_graph = \
                block_building_method.build_blocks(data,
                                            attributes_1=self.block_building["attributes_1"] \
                                                                if "attributes_1" in self.block_building else None,
                                                attributes_2=self.block_building["attributes_2"] \
                                                                if "attributes_2" in self.block_building else None,
                                                tqdm_disable=workflow_step_tqdm_disable)
        else:
            block_building_blocks, em_graph = \
                block_building_method.build_blocks(data, 
                                                   attributes_1=self.block_building["attributes_1"] \
                                                                if "attributes_1" in self.block_building else None,
                                                    attributes_2=self.block_building["attributes_2"] \
                                                                if "attributes_2" in self.block_building else None,
                                                tqdm_disable=workflow_step_tqdm_disable,
                                                with_entity_matching=True,
                                                **self.block_building["exec_params"])                

        self.final_pairs = block_building_blocks
        res = block_building_method.evaluate(block_building_blocks,
                                            export_to_dict=True,
                                            with_classification_report=with_classification_report,
                                            verbose=verbose)
        self._save_step(res, block_building_method.method_configuration())
        self._workflow_bar.update(1)
        #
        # Clustering step [optional]
        #
        if self.clustering:
            clustering_method = self.clustering['method'](**self.clustering["params"]) \
                                            if "params" in self.clustering \
                                            else self.clustering['method']()
            self.final_pairs = components = clustering_method.process(em_graph, data)
            res = clustering_method.evaluate(components,
                                            export_to_dict=True,
                                            with_classification_report=False,
                                            verbose=verbose)
            self._save_step(res, clustering_method.method_configuration())
            self.workflow_exec_time = time() - start_time
            self._workflow_bar.update(1)
        # self.runtime.append(self.workflow_exec_time)

# class SimilarityJoinsWorkFlow(PYJEDAIWorkFlow):
#     raise NotImplementedError("Joins workflow is not implemented yet.")

# class ProgressiveWorkFlow(PYJEDAIWorkFlow):
#     raise NotImplementedError("Progressive workflow is not implemented yet.")


"""Datamodel of pyjedai.
"""
import pandas as pd
from pandas import DataFrame, concat
import re
import nltk
nltk.download('stopwords')
from nltk.corpus import stopwords

from abc import ABC, abstractmethod
from collections import defaultdict
from ordered_set import OrderedSet

class PYJEDAIFeature(ABC):

    _method_name: str
    _method_info: str
    _method_short_name: str

    def __init__(self) -> None:
        super().__init__()
        self._progress_bar: tqdm
        self.execution_time: float
        self.tqdm_disable: bool
        self.data: Data

    @abstractmethod
    def _configuration(self) -> dict:
        pass

    @abstractmethod
    def evaluate(self,
                 prediction=None,
                 export_to_df: bool = False,
                 export_to_dict: bool = False,
                 with_classification_report: bool = False,
                 verbose: bool = True) -> any:
        pass
    
    def method_configuration(self) -> dict:
        """Returns configuration details
        """
        return {
            "name" : self._method_name,
            "parameters" : self._configuration(),
            "runtime": self.execution_time
        }
        
    def report(self) -> None:
        """Prints Block Building method configuration
        """
        parameters = ("\n" + ''.join(['\t{0}: {1}\n'.format(k, v) for k, v in self._configuration().items()])) \
                        if len(self._configuration().items()) != 0 else ' None'
        print(
            "Method name: " + self._method_name +
            "\nMethod info: " + self._method_info +
            "\nParameters: " + parameters +
            "\nRuntime: {:2.4f} seconds".format(self.execution_time)
        )

    @abstractmethod
    def stats(self) -> None:
        pass

class Data:
    """The corpus of the dataset that will be processed with pyjedai. \
        Contains all the information of the dataset and will be passed to each step \
        of the ER workflow.
    """

    def __init__(
                self,
                dataset_1: DataFrame,
                id_column_name_1: str,                
                attributes_1: list = None,
                dataset_name_1: str = None,
                dataset_2: DataFrame = None,
                attributes_2: list = None,
                id_column_name_2: str = None,
                dataset_name_2: str = None,
                ground_truth: DataFrame = None,
    ) -> None:
        # Original Datasets as pd.DataFrame
        if isinstance(dataset_1, pd.DataFrame):
            self.dataset_1 = dataset_1
        else:
            raise AttributeError("Dataset 1 must be a pandas DataFrame")

        if dataset_2 is not None:
            if id_column_name_2 is None:
                raise AttributeError("Must provide datasets 2 id column")

            if isinstance(dataset_2, pd.DataFrame):
                self.dataset_2 = dataset_2
            else:
                raise AttributeError("Dataset 2 must be a pandas DataFrame")

        # Processed dataframes to lists (all attribute columns)
        # Tranformed to list for optimization (list)
        self.entities_d1: list
        self.entities_d2: list = None

        # D1 and D2 dataframes concatenated
        self.entities: DataFrame

        # Datasets specs
        self.is_dirty_er = dataset_2 is None
        self.dataset_limit = self.num_of_entities_1 = len(dataset_1)
        self.num_of_entities_2: int = len(dataset_2) if dataset_2 is not None else 0
        self.num_of_entities: int = self.num_of_entities_1 + self.num_of_entities_2

        self.id_column_name_1 = id_column_name_1
        self.id_column_name_2 = id_column_name_2
        
        self.dataset_name_1 = dataset_name_1
        self.dataset_name_2 = dataset_name_2
        
        # Fill NaN values with empty string
        self.dataset_1.fillna("", inplace=True)
        if not self.is_dirty_er:
            self.dataset_2.fillna("", inplace=True)

        self.dataset_name_1 = dataset_name_1
        self.dataset_name_2 = dataset_name_2
        
        # Fill NaN values with empty string
        self.dataset_1.fillna("", inplace=True)
        if not self.is_dirty_er:
            self.dataset_2.fillna("", inplace=True)
            
        # Attributes
        if attributes_1 is None:
            if dataset_1.columns.values.tolist():
                self.attributes_1 = dataset_1.columns.values.tolist()
                if self.id_column_name_1 in self.attributes_1:
                    self.attributes_1.remove(self.id_column_name_1)
            else:
                raise AttributeError(
                    "Dataset 1 must contain column names if attributes_1 is empty.")
        else:
            self.attributes_1: list = attributes_1

        if dataset_2 is not None:
            if attributes_2 is None:
                if dataset_2.columns.values.tolist():
                    self.attributes_2 = dataset_2.columns.values.tolist()
                    if self.id_column_name_2 in self.attributes_2:
                        self.attributes_2.remove(self.id_column_name_1)
                else:
                    raise AttributeError("Dataset 2 must contain column names if attributes_2 is empty.")
            else:
                self.attributes_2: list = attributes_2

        # Ground truth data
        if ground_truth is not None:
            self.ground_truth = ground_truth.astype(str)
            self._ids_mapping_1: dict
            self._gt_to_ids_reversed_1: dict
            self._ids_mapping_2: dict
            self._gt_to_ids_reversed_2: dict

        self.entities = self.dataset_1 = self.dataset_1.astype(str)
        
        # Concatenated columns into new dataframe
        self.entities_d1 = self.dataset_1[self.attributes_1]

        if not self.is_dirty_er:
            self.dataset_2 = self.dataset_2.astype(str)
            self.entities_d2 = self.dataset_2[self.attributes_2]
            self.entities = pd.concat([self.dataset_1, self.dataset_2],
                                      ignore_index=True)

        if ground_truth is not None:
            self._create_gt_mapping()
            self._store_pairs()
        else:
            self.ground_truth = None

    # def _store_pairs(self) -> None:
    #     """Creates a mapping:
    #         - pairs_of : ids of first dataset to ids of true matches from second dataset"""
        
    #     self.pairs_of = defaultdict(set)
    #     d1_col_index, d2_col_index = (0, 1) if self.inorder_gt else (1,0)
        
    #     for _, row in self.ground_truth.iterrows():
    #         id1, id2 = (row[d1_col_index], row[d2_col_index])
    #         if id1 in self.pairs_of: self.pairs_of[id1].append(id2)
    #         else: self.pairs_of[id1] = [id2]  
    
    
    def _store_pairs(self) -> None:
        """Creates a mapping:
            - pairs_of : ids of first dataset to ids of true matches from second dataset"""
        
        self.duplicate_of = defaultdict(set)
        
        for _, row in self.ground_truth.iterrows():
            id1, id2 = (row[0], row[1])
            if id1 in self.duplicate_of: self.duplicate_of[id1].add(id2)
            else: self.duplicate_of[id1] = {id2}
            
    def _are_true_positives(self, id1 : int, id2 : int):
        """Checks if given pair of identifiers represents a duplicate.
           Identifiers must be inorder, first one belonging to the first and the second to the second dataset

        Args:
            id1 (int, optional): Identifier from the first dataframe. 
            id2 (int, optional): Identifier from the second dataframe.

        Returns:
            _type_: _description_
        """
        return id1 in self.duplicate_of and id2 in self.duplicate_of[id1]
    
    def _create_gt_mapping(self) -> None:
        """Creates two mappings:
            - _ids_mapping_X: ids from initial dataset to index
            - _gt_to_ids_reversed_X (inversed _ids_mapping_X): index number \
                            from range to initial dataset id
        """
        if self.ground_truth is not None:
            self.ground_truth = self.ground_truth.astype(str)
        else:
            return

        self._ids_mapping_1 = dict(
            zip(
                self.dataset_1[self.id_column_name_1].tolist(),
                range(0, self.num_of_entities_1)
            )
        )

        self._gt_to_ids_reversed_1 = dict(
            zip(
                self._ids_mapping_1.values(),
                self._ids_mapping_1.keys()
            )
        )

        if not self.is_dirty_er:
            self._ids_mapping_2 = dict(
                zip(
                    self.dataset_2[self.id_column_name_2].tolist(),
                    range(self.num_of_entities_1, self.num_of_entities_1+self.num_of_entities_2)
                )
            )

            self._gt_to_ids_reversed_2 = dict(
                zip(
                    self._ids_mapping_2.values(),
                    self._ids_mapping_2.keys()
                )
            )

    def print_specs(self) -> None:
        """Dataset report.
        """
        print(25*"-", "Data", 25*"-")
        print("Type of Entity Resolution: ", "Dirty" if self.is_dirty_er else "Clean-Clean" )
        print("Dataset-1:")
        print("\tNumber of entities: ", self.num_of_entities_1)
        print("\tNumber of NaN values: ", self.dataset_1.isnull().sum().sum())
        print("\tAttributes: \n\t\t", self.attributes_1)
        if not self.is_dirty_er:
            print("Dataset-2:")
            print("\tNumber of entities: ", self.num_of_entities_2)
            print("\tNumber of NaN values: ", self.dataset_2.isnull().sum().sum())
            print("\tAttributes: \n\t\t", self.attributes_2)
        print("\nTotal number of entities: ", self.num_of_entities)
        if self.ground_truth is not None:
            print("Number of matching pairs in ground-truth: ", len(self.ground_truth))
        print(56*"-", "\n")

    
    # Functions that removes stopwords, punctuation, uni-codes, numbers from the dataset
    def clean_dataset(self, 
                      remove_stopwords: bool = True, 
                      remove_punctuation: bool = True, 
                      remove_numbers:bool = True,
                      remove_unicodes: bool = True) -> None:
        """Removes stopwords, punctuation, uni-codes, numbers from the dataset.
        """
        
        # Make self.dataset_1 and self.dataset_2 lowercase
        self.dataset_1 = self.dataset_1.applymap(lambda x: x.lower())
        if not self.is_dirty_er:
            self.dataset_2 = self.dataset_2.applymap(lambda x: x.lower())
            
        if remove_numbers:
            self.dataset_1 = self.dataset_1.applymap(lambda x: re.sub(r'\d+', '', x))
            if not self.is_dirty_er:
                self.dataset_2 = self.dataset_2.applymap(lambda x: re.sub(r'\d+', '', x))    
                
        if remove_unicodes:
            self.dataset_1 = self.dataset_1.applymap(lambda x: re.sub(r'[^\x00-\x7F]+', '', x))
            if not self.is_dirty_er:
                self.dataset_2 = self.dataset_2.applymap(lambda x: re.sub(r'[^\x00-\x7F]+', '', x))
            
        if remove_punctuation:
            self.dataset_1  = self.dataset_1.applymap(lambda x: re.sub(r'[^\w\s]','',x))
            if not self.is_dirty_er:
                self.dataset_2 = self.dataset_2.applymap(lambda x: re.sub(r'[^\w\s]','',x))
        
        if remove_stopwords:
            self.dataset_1 = self.dataset_1.applymap(lambda x: ' '.join([word for word in x.split() if word not in (stopwords.words('english'))]))
            if not self.is_dirty_er:
                self.dataset_2 = self.dataset_2.applymap(lambda x: ' '.join([word for word in x.split() if word not in (stopwords.words('english'))]))    

        self.entities = self.dataset_1 = self.dataset_1.astype(str)
        
        # Concatenated columns into new dataframe
        self.entities_d1 = self.dataset_1[self.attributes_1]

        if not self.is_dirty_er:
            self.dataset_2 = self.dataset_2.astype(str)
            self.entities_d2 = self.dataset_2[self.attributes_2]
            self.entities = pd.concat([self.dataset_1, self.dataset_2],
                                      ignore_index=True)

    def stats_about_data(self) -> None:
        
        stats_df = pd.DataFrame(columns=['word_count_1', 'word_count_2'])
        
        # Calculate the average number of words per line
        stats_df['word_count_1'] = self.dataset_1.apply(lambda row: len(row.str.split()), axis=1)
        print(stats_df['word_count_1'])
        average_words_per_line_1 = stats_df['word_count_1'].mean()
        print(average_words_per_line_1)
        
        if not self.is_dirty_er:
            stats_df['word_count_2'] = self.dataset_2.apply(lambda row: len(row.str.split()), axis=1)
            average_words_per_line_2 = stats_df['word_count_2'].mean()
            print(average_words_per_line_2)
            
        return stats_df

        
class Block:
    """The main module used for storing entities in the blocking steps of pyjedai module. \
        Consists of 2 sets of profile entities 1 for Dirty ER and 2 for Clean-Clean ER.
    """
    def __init__(self) -> None:
        self.entities_D1: set = OrderedSet()
        self.entities_D2: set = OrderedSet()

    def get_cardinality(self, is_dirty_er) -> int:
        """Returns block cardinality.

        Args:
            is_dirty_er (bool): Dirty or Clean-Clean ER.

        Returns:
            int: Cardinality
        """
        if is_dirty_er:
            return len(self.entities_D1)*(len(self.entities_D1)-1)/2
        return len(self.entities_D1) * len(self.entities_D2)

    def get_size(self) -> int:
        """Returns block size.

        Returns:
            int: Block size
        """
        return len(self.entities_D1) + len(self.entities_D2)

    def verbose(self, key: any, is_dirty_er: bool) -> None:
        """Prints contents of a block

        Args:
            key (any): Block key
            is_dirty_er (bool): Dirty or Clean-Clean ER.
        """
        print("\nBlock ", "\033[1;32m"+key+"\033[0m", " has cardinality ", str(self.get_cardinality(is_dirty_er)) ," and contains entities with ids: ")
        if is_dirty_er:
            print("Dirty dataset: " + "[\033[1;34m" + \
             str(len(self.entities_D1)) + " entities\033[0m]")
            print(self.entities_D1)
        else:
            print("Clean dataset 1: " + "[\033[1;34m" + \
             str(len(self.entities_D1)) + " entities\033[0m]")
            print(self.entities_D1)
            print("Clean dataset 2: " + "[\033[1;34m" + str(len(self.entities_D2)) + \
            " entities\033[0m]")
            print(self.entities_D2)



import itertools 
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import os
from typing import List, Tuple

# Function that creates a confusion matrix
def create_confusion_matrix(confusion_matrix, title):
    
    plt.figure(figsize = (8,5))
    classes = ['Different','Matching']
    cmap = plt.cm.Blues
    plt.grid(False)
    plt.imshow(confusion_matrix, interpolation='nearest', cmap=cmap)
    plt.title(title)
    plt.colorbar()
    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=45)
    plt.yticks(tick_marks, classes)

    thresh = confusion_matrix.max() / 2.
    for i, j in itertools.product(range(confusion_matrix.shape[0]), range(confusion_matrix.shape[1])):
        plt.text(j, i, confusion_matrix[i, j],horizontalalignment="center",color="white" if confusion_matrix[i, j] > thresh else "black")

    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel('Predicted label')
    plt.tight_layout()
    plt.ylim([1.5, -.5])
    plt.show() 

def plot_feature_progress_per_attribute_group(method_name : str,
                                              dataset_name : str,
                                              feature : str,
                                              attributes : list,
                                              df : pd.DataFrame = None,
                                              load_path : str = None,
                                              grid : bool = True,
                                              save : bool = True,
                                              verbose : bool = True,
                                              in_plot_directory : bool = True
                                              ) -> None:
    """Plots the progress of the value of requested feature per budget for experiments grouped by the attributes.
       Saves the plot as an image in the requested path.
    
    Args:
        method_name (str): Name of the method used in the dataframe's experiments
        dataset_name (str): Name of dataset on which the dataframe's experiments have been applied on
        feature (str): The feature whose per budget progress we want to plot (e.x. auc)
        attributes (list): Group of experiments' arguments whose each distinct combination constitutes a seperate curve
        df (pd.DataFrame): Dataframe containing the information about progressive PER experiments (Defaults to None)
        load_path (str): Path from which the dataframe should be loaded from (Defaults to None)
        grid (bool): Grid to be displayed in the plot (Defaults to True)
        save (bool) : Save the plot as an image on disk (Defaults to True)
        verbose (bool) : Show the produced plot
        in_plot_directory (bool) : Plot to be saved in an experiment directory - 
        created in the target dataframe's / current directory if non-existent (Defaults to True)
    """
    
    experiments : pd.DataFrame
    if(df is not None):
        experiments = df
    elif(load_path is not None):
        experiments = pd.read_csv(load_path)
    else:
        raise ValueError("No dataframe or csv file given - Cannot plot the experiments.")
    
    experiments = experiments.groupby(attributes)


    fig = plt.figure(figsize=(16, 12))
    ax = plt.subplot(111)

    for attributes_unique_values, attributes_experiment_group in experiments:
        group_name = '-'.join([str(attribute) for attribute in attributes_unique_values])
        attributes_experiment_group_per_budget = attributes_experiment_group.sort_values(by='budget').groupby('budget')
        budgets = []
        average_feature_values = []
        for _, current_budget_attributes_experiment_group in attributes_experiment_group_per_budget:
            budgets.append(current_budget_attributes_experiment_group['budget'].mean())
            average_feature_values.append(current_budget_attributes_experiment_group[feature].mean())

        ax.plot(budgets, average_feature_values, label=str(group_name), marker='o', linestyle='-')

    # Customize the plot
    ax.set_title(f'{method_name.capitalize()}/{dataset_name.capitalize()} - Average {feature.capitalize()} vs. Budget Curves')
    ax.set_xlabel('Budget')
    ax.set_ylabel(f'Average {feature.capitalize()}')
    
    pos = ax.get_position()
    ax.set_position([pos.x0, pos.y0, pos.width * 0.9, pos.height])
    ax.legend(title=attributes, fontsize="9", loc='center right', bbox_to_anchor=(1.23, 0.5))
    
    ax.grid(grid)
    
    if(save):
        file_name = '_'.join([dataset_name, method_name, feature, 'for', '_'.join(attributes)]) + '.png'
        dataframe_directory = os.path.dirname(load_path) if load_path is not None else './'
        store_directory = dataframe_directory if not in_plot_directory else os.path.join(dataframe_directory, 'plots/')        
        
        if in_plot_directory and not os.path.exists(store_directory):
            os.makedirs(store_directory)
            
        plt.savefig(os.path.join(store_directory, file_name))
        
    plt.show()
    
    
    
def plot_attribute_group_avg_ranking(method_name : str,
                                     feature : str,
                                     attributes : list,
                                     dfs : List[pd.DataFrame] = None,
                                     load_paths : List[str] = None,
                                     grid : bool = True,
                                     save : bool = True,
                                     verbose : bool = True,
                                     in_plot_directory : bool = True
                                     ) -> None:
    """For each unique combination of given attributes calculates its average feature value across datasets for each budget.
       Plots the corresponding results and stores them as an image if it is requested.
    
    Args:
        method_name (str): The name of the PER method whose experiments we are evaluating
        feature (str): The feature that we want to evaluate the average ranking of the attribute group for  
        attributes (list): Group of experiments' arguments whose each distinct combination constitutes a seperate curve
        dfs (List[pd.DataFrame]): Dataframes containing the information about progressive PER experiments (Defaults to None)
        load_paths (List[str]): Paths from which the dataframe should be loaded from (Defaults to None)
        grid (bool): Grid to be displayed in the plot (Defaults to True)
        save (bool) : Save the plot as an image on disk (Defaults to True)
        verbose (bool) : Show the produced plot
        in_plot_directory (bool) : Plot to be saved in an experiment directory - 
        created in the target dataframe's / current directory if non-existent (Defaults to True)
    """
    
    if(dfs is None and load_paths is None):
        raise ValueError("No dataframes or csv files given - Cannot calculate and plot average combinations rankings.")

    total_datasets = len(dfs) if dfs is not None else len(load_paths)
    attributes_combinations = {}
    attributes_combinations_budget_scores : List[Tuple[float, str]]
    
    for current_dataset in range(total_datasets):
        if(dfs is not None):
            experiments = dfs[current_dataset]
        else:
            current_dataset_path = load_paths[current_dataset]
            experiments = pd.read_csv(current_dataset_path)
    
        budgets_experiments = experiments.sort_values(by='budget').groupby('budget')

        for current_budget, current_budget_experiments in budgets_experiments:
            current_budget_attributes_combinations = current_budget_experiments.groupby(attributes[0] if len(attributes) == 1 else attributes)
            attributes_combinations_budget_scores = []
            
            for attributes_combination, current_budget_attributes_combination in current_budget_attributes_combinations:
                attributes_combination_budget_feature_value = current_budget_attributes_combination[feature].mean()
                attributes_combinations_budget_scores.append((attributes_combination_budget_feature_value, attributes_combination))
                
            for ranking, attributes_combinations_budget_score in enumerate(sorted(attributes_combinations_budget_scores, reverse=True)):
                attributes_combination_budget_feature_value, attributes_combination = attributes_combinations_budget_score
                if attributes_combination not in attributes_combinations:
                    attributes_combinations[attributes_combination] = {}
                    
                if current_budget not in attributes_combinations[attributes_combination]:
                    attributes_combinations[attributes_combination][current_budget] = []
                    
                attributes_combinations[attributes_combination][current_budget].append(ranking+1)

    fig = plt.figure(figsize=(16, 12))
    ax = plt.subplot(111)       
            
    for attributes_combination, attributes_combination_budgets in attributes_combinations.items():
        
        attributes_combination_average_rankings = []
        sorted_budgets = sorted(attributes_combination_budgets.keys(), reverse=False)
        for budget in sorted_budgets:
            attributes_combination_average_rankings.append(sum(attributes_combination_budgets[budget]) / len(attributes_combination_budgets[budget]))
            
        ax.plot(sorted_budgets, attributes_combination_average_rankings, label=str(attributes_combination), marker='o', linestyle='-')
        
        
    # Customize the plot
    ax.set_title(f'{method_name.capitalize()} - Average {feature.capitalize()} Ranking vs. Budget Curves')
    ax.set_xlabel('Budget')
    ax.set_ylabel('Average Ranking')
    
    pos = ax.get_position()
    ax.set_position([pos.x0, pos.y0, pos.width * 0.9, pos.height])
    ax.legend(title=attributes, fontsize="9", loc='center right', bbox_to_anchor=(1.23, 0.5))
    
    ax.grid(grid)
    
    if(save):
        file_name = '_'.join([method_name, 'for', '_'.join(attributes), 'avg_rankings', feature]) + '.png'
        dataframe_directory = os.path.dirname(load_paths[0]) if load_paths is not None else './'
        store_directory = dataframe_directory if not in_plot_directory else os.path.join(dataframe_directory, 'avg_rankings/')        
        
        if in_plot_directory and not os.path.exists(store_directory):
            os.makedirs(store_directory)
            
        plt.savefig(os.path.join(store_directory, file_name))
        
    plt.show()
    
    
    
def plot_attribute_group_avg_top_distance(method_name : str,
                                     feature : str,
                                     attributes : list,
                                     dfs : List[pd.DataFrame] = None,
                                     load_paths : List[str] = None,
                                     grid : bool = True,
                                     save : bool = True,
                                     verbose : bool = True,
                                     in_plot_directory : bool = True
                                     ) -> None:
    """For each unique combination of given attributes calculates its feature's value average difference from the best value across datasets for each budget.
       Plots the corresponding results and stores them as an image if it is requested.
    
    Args:
        method_name (str): The name of the PER method whose experiments we are evaluating
        feature (str): The feature that we want to evaluate the average ranking of the attribute group for  
        attributes (list): Group of experiments' arguments whose each distinct combination constitutes a seperate curve
        dfs (List[pd.DataFrame]): Dataframes containing the information about progressive PER experiments (Defaults to None)
        load_paths (List[str]): Paths from which the dataframe should be loaded from (Defaults to None)
        grid (bool): Grid to be displayed in the plot (Defaults to True)
        save (bool) : Save the plot as an image on disk (Defaults to True)
        verbose (bool) : Show the produced plot
        in_plot_directory (bool) : Plot to be saved in an experiment directory - 
        created in the target dataframe's / current directory if non-existent (Defaults to True)
    """
    
    if(dfs is None and load_paths is None):
        raise ValueError("No dataframes or csv files given - Cannot calculate and plot average combinations rankings.")

    total_datasets = len(dfs) if dfs is not None else len(load_paths)
    attributes_combinations = {}
    attributes_combinations_budget_scores : List[Tuple[float, str]]
    
    for current_dataset in range(total_datasets):
        if(dfs is not None):
            experiments = dfs[current_dataset]
        else:
            current_dataset_path = load_paths[current_dataset]
            experiments = pd.read_csv(current_dataset_path)
    
        budgets_experiments = experiments.sort_values(by='budget').groupby('budget')

        for current_budget, current_budget_experiments in budgets_experiments:
            current_budget_attributes_combinations = current_budget_experiments.groupby(attributes[0] if len(attributes) == 1 else attributes)
            attributes_combinations_budget_scores = []
            
            for attributes_combination, current_budget_attributes_combination in current_budget_attributes_combinations:
                attributes_combination_budget_feature_value = current_budget_attributes_combination[feature].mean()
                attributes_combinations_budget_scores.append((attributes_combination_budget_feature_value, attributes_combination))
                
            attributes_combinations_budget_scores = sorted(attributes_combinations_budget_scores, reverse=True)
            budget_highest_feature_value = attributes_combinations_budget_scores[0][0]
                
            for attributes_combinations_budget_score in attributes_combinations_budget_scores:
                attributes_combination_budget_feature_value, attributes_combination = attributes_combinations_budget_score
                if attributes_combination not in attributes_combinations:
                    attributes_combinations[attributes_combination] = {}
                    
                if current_budget not in attributes_combinations[attributes_combination]:
                    attributes_combinations[attributes_combination][current_budget] = []
                    
                attributes_combinations[attributes_combination][current_budget].append(budget_highest_feature_value - attributes_combination_budget_feature_value)

    fig = plt.figure(figsize=(16, 12))
    ax = plt.subplot(111)       
            
    for attributes_combination, attributes_combination_budgets in attributes_combinations.items():
        
        attributes_combination_average_rankings = []
        sorted_budgets = sorted(attributes_combination_budgets.keys(), reverse=False)
        for budget in sorted_budgets:
            attributes_combination_average_rankings.append(sum(attributes_combination_budgets[budget]) / len(attributes_combination_budgets[budget]))
            
        ax.plot(sorted_budgets, attributes_combination_average_rankings, label=str(attributes_combination), marker='o', linestyle='-')
        
        
    # Customize the plot
    ax.set_title(f'{method_name.capitalize()} - Average {feature.capitalize()} Distance from Top vs. Budget Curves')
    ax.set_xlabel('Budget')
    ax.set_ylabel('Average Distance from Top')
    
    pos = ax.get_position()
    ax.set_position([pos.x0, pos.y0, pos.width * 0.9, pos.height])
    ax.legend(title=attributes, fontsize="9", loc='center right', bbox_to_anchor=(1.23, 0.5))
    
    ax.grid(grid)
    
    if(save):
        file_name = '_'.join([method_name, 'for', '_'.join(attributes), 'avg_distances', feature]) + '.png'
        dataframe_directory = os.path.dirname(load_paths[0]) if load_paths is not None else './'
        store_directory = dataframe_directory if not in_plot_directory else os.path.join(dataframe_directory, 'avg_distances/')        
        
        if in_plot_directory and not os.path.exists(store_directory):
            os.makedirs(store_directory)
            
        plt.savefig(os.path.join(store_directory, file_name))
        
    plt.show()
    
    
def plot_attributes_performance_for_budget(method_name : str,
                                            feature : str,
                                            attributes : list,
                                            dfs : List[pd.DataFrame] = None,
                                            load_paths : List[str] = None,
                                            calculate_distance : bool = False,
                                            grid : bool = True,
                                            save : bool = True,
                                            verbose : bool = True,
                                            in_plot_directory : bool = True
                                            ) -> pd.DataFrame:
    """For each unique combination of given attributes calculates its feature value's average distance from best / ranking per budget.
       Then calculates the same values for each combination of budget and dataset. Combination rows are sorted by the average of the averages
       of the feature value's distance from best / ranking per budget.
    
    Args:
        method_name (str): The name of the PER method whose experiments we are evaluating
        feature (str): The feature that we want to evaluate the average ranking of the attribute group for  
        attributes (list): Group of experiments' arguments whose each distinct combination constitutes a seperate curve
        dfs (List[pd.DataFrame]): Dataframes containing the information about progressive PER experiments (Defaults to None)
        load_paths (List[str]): Paths from which the dataframe should be loaded from (Defaults to None)
        calculate_distance (bool): Calculate distance for the feature from top within dataset (Defaults to False)
        grid (bool): Grid to be displayed in the plot (Defaults to True)
        save (bool) : Save the plot as an image on disk (Defaults to True)
        verbose (bool) : Show the produced plot
        in_plot_directory (bool) : Plot to be saved in an experiment directory - 
        created in the target dataframe's / current directory if non-existent (Defaults to True)
    Returns:
        pd.DataFrame: Dataframe containing the performance of the feature for each attributes' value combination across all datasets
                      for the requested budget order (e.x. first budget for each dataset)
    """
    
    if(dfs is None and load_paths is None):
        raise ValueError("No dataframes or csv files given - Cannot calculate and plot average combinations rankings.")
            
    total_datasets : int = len(dfs) if dfs is not None else len(load_paths)
    attributes_combinations : dict = {}  
    budget_dataset_best_feature_value : dict = {}
    
    attributes_column : str = ' + '.join([' '.join([word.capitalize() for word in attribute.split('_')]) for attribute in attributes])
    budget_dataframe : dict = {attributes_column : []}    

    for current_dataset in range(total_datasets):
        if(dfs is not None):
            experiments = dfs[current_dataset]
        else:
            current_dataset_path = load_paths[current_dataset]
            experiments = pd.read_csv(current_dataset_path)
            
        current_dataset_name : str = "D" + str(current_dataset+1)
        budgets_experiments = experiments.sort_values(by='budget').groupby('budget')
        
        current_budget = 0
        for _, current_budget_experiments in budgets_experiments:
            current_budget += 1
            current_budget_attributes_combinations = current_budget_experiments.groupby(attributes[0] if len(attributes) == 1 else attributes)

            for attributes_combination, current_budget_attributes_combination in current_budget_attributes_combinations:
                
                if attributes_combination not in attributes_combinations:
                    attributes_combinations[attributes_combination] = {}
                    
                if current_budget not in attributes_combinations[attributes_combination]:
                    attributes_combinations[attributes_combination][current_budget] = {}
                current_budget_attributes_combination_feature_value = current_budget_attributes_combination[feature].mean()          
                attributes_combinations[attributes_combination][current_budget][current_dataset_name] = current_budget_attributes_combination_feature_value
                
                if current_budget not in budget_dataset_best_feature_value:
                    budget_dataset_best_feature_value[current_budget] = {}
                    
                if current_dataset_name not in budget_dataset_best_feature_value[current_budget]:
                    budget_dataset_best_feature_value[current_budget][current_dataset_name] = 0.0
                    
                if(current_budget_attributes_combination_feature_value > budget_dataset_best_feature_value[current_budget][current_dataset_name]):
                    budget_dataset_best_feature_value[current_budget][current_dataset_name] = current_budget_attributes_combination_feature_value
                    
            if calculate_distance:
                # we want to calculate each combination's performance distance from best performance per dataset
                for attributes_combination, current_budget_attributes_combination in current_budget_attributes_combinations:
                    attributes_combinations[attributes_combination][current_budget][current_dataset_name] = budget_dataset_best_feature_value[current_budget][current_dataset_name] - attributes_combinations[attributes_combination][current_budget][current_dataset_name]
            else:
                # we want to calculate each combination's ranking per dataset
                combinations_performance : List[Tuple[float, str]] = []
                for attributes_combination, current_budget_attributes_combination in current_budget_attributes_combinations:
                    combinations_performance.append((attributes_combinations[attributes_combination][current_budget][current_dataset_name], attributes_combination))
                combinations_performance = sorted(combinations_performance, reverse=True)
                
                for ranking, combination_performance in enumerate(combinations_performance):
                    performance, combination = combination_performance
                    attributes_combinations[combination][current_budget][current_dataset_name] = ranking + 1
                    
                    
    for attributes_combination, budgets_attributes_combination in attributes_combinations.items(): 
        budget_dataframe[attributes_column].append(attributes_combination)
        for budget in budgets_attributes_combination:
            
            budget_attribute_combination = budgets_attributes_combination[budget]
            budget_name = "B" + str(budget)
            budget_feature_avg_value = 0.0
            
            for dataset, dataset_budget_attribute_combination in budget_attribute_combination.items():
                
                budget_dataset_column = '_'.join([str(budget_name),str(dataset)])
                
                if(budget_dataset_column not in budget_dataframe):
                    budget_dataframe[budget_dataset_column] = []
                
                budget_dataset_feature_value = attributes_combinations[attributes_combination][budget][dataset]
                budget_feature_avg_value += budget_dataset_feature_value   
                budget_dataframe[budget_dataset_column].append(budget_dataset_feature_value)
                
            budget_average_column = '_'.join(["AVERAGE",budget_name])
            if(budget_average_column not in budget_dataframe):
                    budget_dataframe[budget_average_column] = []
                    
            budget_dataframe[budget_average_column].append(budget_feature_avg_value / len(budget_attribute_combination))
     
    budget_dataframe = pd.DataFrame(budget_dataframe)       
    # Sort Attributes Combinations rows based on the average of the averages of their per budget performances
    average_budget_performance_columns = ['_'.join(["AVERAGE", "B" + str(index+1)]) for index in range(len(budgets_experiments))]
    budget_dataframe['AA_BS'] = budget_dataframe[average_budget_performance_columns].mean(axis=1)
    budget_dataframe = budget_dataframe.sort_values(by='AA_BS', ascending=True)
    
    if(save):
        metric = "distance" if calculate_distance else "ranking"
        file_name = '_'.join([feature, metric, 'for', method_name, 'with', '_'.join(attributes)]) + '.csv'
        
        dataframe_directory = os.path.dirname(load_paths[0]) if load_paths is not None else './'
        store_directory = dataframe_directory if not in_plot_directory else os.path.join(dataframe_directory, metric + '-analytical-performances/')        
        
        if in_plot_directory and not os.path.exists(store_directory):
            os.makedirs(store_directory)
            
        budget_dataframe.to_csv(os.path.join(store_directory, file_name), index=False)
            
    return budget_dataframe    
                     
    # total_datasets : int = len(dfs) if dfs is not None else len(load_paths)
    # attributes_combinations : dict = {}  
    # budget_dataset_best_feature_value : dict = {}
    # attributes_column : str = '_'.join([attribute.capitalize() for attribute in attributes])
    
    # budget_dataframe : dict = {attributes_column : [], "AVERAGE" : []}
    # budget_dataframe_column_data_types = {"AVERAGE" : float}

    
    # for current_dataset in range(total_datasets):
    #     if(dfs is not None):
    #         experiments = dfs[current_dataset]
    #     else:
    #         current_dataset_path = load_paths[current_dataset]
    #         experiments = pd.read_csv(current_dataset_path)
        
    #     current_dataset_name : str = "D" + str(current_dataset+1)    
    #     budget_dataframe_column_data_types[current_dataset_name] = float if calculate_distance else int
        
    #     budget_dataset_best_feature_value[current_dataset_name] = {}
    #     budget_dataframe[current_dataset_name] = []
        
        
    #     budgets_experiments = experiments.sort_values(by='budget').groupby('budget')
        
    #     for current_budget, current_budget_experiments in budgets_experiments:
        
    #         current_budget_experiments = budgets_experiments.get_group(list(budgets_experiments.groups.keys())[budget_order])
    #         current_budget_attributes_combinations = current_budget_experiments.groupby(attributes[0] if len(attributes) == 1 else attributes)
            
    #         for attributes_combination, current_budget_attributes_combination in current_budget_attributes_combinations:
                
    #             if attributes_combination not in attributes_combinations:
    #                 attributes_combinations[attributes_combination] = {}
                    
    #             if current_budget not in attributes_combinations[attributes_combination]:
    #                 attributes_combinations[attributes_combination][current_budget] = {}
                    
    #             current_budget_attributes_combination_feature_value = current_budget_attributes_combination[feature].mean()           
    #             attributes_combinations[attributes_combination][current_budget][current_dataset] = current_budget_attributes_combination_feature_value
                
                
                
    #             if current_budget_attributes_combination_feature_value > dataset_best_feature_value[current_dataset_name]:
    #                 dataset_best_feature_value[current_dataset_name] = current_budget_attributes_combination_feature_value
                        
    #         if calculate_distance:
    #             # we want to calculate each combination's performance distance from best performance per dataset
    #             for attributes_combination, current_budget_attributes_combination in current_budget_attributes_combinations:
    #                 attributes_combinations[attributes_combination][current_dataset_name] = dataset_best_feature_value[current_dataset_name] - attributes_combinations[attributes_combination][current_dataset_name]
    #         else:
    #             # we want to calculate each combination's ranking per dataset
    #             combinations_performance : List[Tuple[float, str]] = []
    #             for attributes_combination, current_budget_attributes_combination in current_budget_attributes_combinations:
    #                 combinations_performance.append((attributes_combinations[attributes_combination][current_dataset_name], attributes_combination))
    #             combinations_performance = sorted(combinations_performance, reverse=True)
                
    #             for ranking, combination_performance in enumerate(combinations_performance):
    #                 performance, combination = combination_performance
    #                 attributes_combinations[combination][current_dataset_name] = ranking + 1
                  
    # for attributes_combination, attributes_combination_datasets_performance in attributes_combinations.items():
        
    #     budget_dataframe[attributes_column].append(attributes_combination)
    #     average_attributes_combination_performance = 0.0
        
    #     for dataset, performance in attributes_combination_datasets_performance.items():
    #         # print(performance)
    #         budget_dataframe[dataset].append(performance)
    #         average_attributes_combination_performance += performance
            
    #     average_attributes_combination_performance /= len(attributes_combination_datasets_performance) 
    #     budget_dataframe["AVERAGE"].append(average_attributes_combination_performance)
        
    # budget_dataframe = pd.DataFrame(budget_dataframe)
    # budget_dataframe = budget_dataframe.astype(budget_dataframe_column_data_types)
    # budget_dataframe = budget_dataframe.sort_values(by='AVERAGE', ascending=True)

    # if(save):
    #     metric = "distance" if calculate_distance else "ranking"
    #     budget_index = "b" + str((budget_order + 1)) 
    #     file_name = '_'.join([budget_index, feature, metric, 'for', method_name, 'with', '_'.join(attributes)]) + '.csv'
        
    #     dataframe_directory = os.path.dirname(load_paths[0]) if load_paths is not None else './'
    #     store_directory = dataframe_directory if not in_plot_directory else os.path.join(dataframe_directory, metric + '-performances/')        
        
    #     if in_plot_directory and not os.path.exists(store_directory):
    #         os.makedirs(store_directory)
            
    #     budget_dataframe.to_csv(os.path.join(store_directory, file_name), index=False)
            
    # return budget_dataframe            



"""Entity Matching Module
"""
import statistics
import pandas as pd
from time import time

import matplotlib.pyplot as plt
import numpy as np
from networkx import Graph
from py_stringmatching.similarity_measure.cosine import Cosine
from py_stringmatching.similarity_measure.dice import Dice
from py_stringmatching.similarity_measure.generalized_jaccard import \
    GeneralizedJaccard
from py_stringmatching.similarity_measure.jaccard import Jaccard
from py_stringmatching.similarity_measure.jaro import Jaro
from py_stringmatching.similarity_measure.levenshtein import Levenshtein
from py_stringmatching.similarity_measure.overlap_coefficient import \
    OverlapCoefficient
from py_stringmatching.tokenizer.qgram_tokenizer import QgramTokenizer
from py_stringmatching.tokenizer.whitespace_tokenizer import \
    WhitespaceTokenizer
from tqdm.autonotebook import tqdm

from .datamodel import Data, PYJEDAIFeature
from .evaluation import Evaluation
from .utils import WordQgramTokenizer, cosine, get_qgram_from_tokenizer_name, FrequencyEvaluator


metrics_mapping = {
    'edit_distance': Levenshtein(),
    'cosine' : Cosine(),
    'jaro' : Jaro(),
    'jaccard' : Jaccard(),
    'generalized_jaccard' : GeneralizedJaccard(),
    'dice': Dice(),
    'overlap_coefficient' : OverlapCoefficient(),
}

vector_metrics_mapping = {
    'cosine': cosine
}

string_metrics = [
    'jaro', 'edit_distance'
]

set_metrics = [
    'cosine', 'dice', 'generalized_jaccard', 'jaccard', 'overlap_coefficient'
]

vector_metrics = [ 
    'cosine', 'dice', 'jaccard'
]

whoosh_index_metrics = [
    'TF-IDF', 'Frequency', 'PL2', 'BM25F'
] 

faiss_metrics = [
    'cosine', 'euclidean'
]

magellan_metrics = string_metrics + set_metrics
available_metrics = magellan_metrics + vector_metrics + whoosh_index_metrics + faiss_metrics

#
# Tokenizers
#
# char_qgram_tokenizers = { 'char_'+ str(i) + 'gram':i for i in range(1, 7) }
# word_qgram_tokenizers = { 'word_'+ str(i) + 'gram':i for i in range(1, 7) }
char_qgram_tokenizers = ['char_tokenizer']
word_qgram_tokenizers = ['word_tokenizer']
magellan_tokenizers = ['white_space_tokenizer']
joins_tokenizers = ["qgrams", "standard", "standard_multiset", "qgrams_multiset"]

# tfidf_tokenizers = [ 'tfidf_' + cq for cq in char_qgram_tokenizers.keys() ] + \
#                     [ 'tfidf_' + wq for wq in word_qgram_tokenizers.keys() ]

# tf_tokenizers = [ 'tf_' + cq for cq in char_qgram_tokenizers.keys() ] + \
#                     [ 'tf_' + wq for wq in word_qgram_tokenizers.keys() ]
                        
# boolean_tokenizers = [ 'boolean_' + cq for cq in char_qgram_tokenizers.keys() ] + \
#                         [ 'boolean_' + wq for wq in word_qgram_tokenizers.keys() ]

# vector_tokenizers = tfidf_tokenizers + tf_tokenizers + boolean_tokenizers

# available_tokenizers = [key for key in char_qgram_tokenizers] + [key for key in word_qgram_tokenizers] + magellan_tokenizers + vector_tokenizers
available_tokenizers = char_qgram_tokenizers + word_qgram_tokenizers + magellan_tokenizers + joins_tokenizers
available_vectorizers = ['tfidf', 'tf', 'boolean']

class AbstractEntityMatching(PYJEDAIFeature):
    """Calculates similarity from 0.0 to 1.0
    """

    _method_name: str = "Abstract Entity Matching"
    _method_info: str = "Calculates similarity from 0. to 1."

    def __init__(
            self,
            metric: str = 'dice',
            similarity_threshold: float = 0.5,
        ) -> None:
        self.pairs: Graph
        self.metric = metric
        self.similarity_threshold = similarity_threshold
        self.execution_time = 0
        self.qgram=None
        #
        # Selecting tokenizer
        #
        if metric not in available_metrics:
            raise AttributeError(
                'Metric ({}) does not exist. Please select one of the available. ({})'.format(
                    metric, available_metrics
                )
            )
        else:
            self._metric = metric

    def _predict_candidate_pairs(self, blocks: dict) -> None:
        """Similarity evaluation after comparison cleaning.

        Args:
            blocks (dict): Comparison cleaning blocks.
        """
        for entity_id, candidates in blocks.items():
            for candidate_id in candidates:
                similarity = self._similarity(entity_id, candidate_id)
                self._insert_to_graph(entity_id, candidate_id, similarity)
            self._progress_bar.update(1)

    def _insert_to_graph(self, entity_id1, entity_id2, similarity):
        if self.similarity_threshold is None or \
            (self.similarity_threshold is not None and similarity > self.similarity_threshold):
            self.pairs.add_edge(entity_id1, entity_id2, weight=similarity)

    def report(self) -> None:
        """Prints Block Building method configuration
        """
        print(
            "Method name: " + self._method_name +
            "\nMethod info: " + self._method_info +
            ("\nParameters: \n" + ''.join(['\t{0}: {1}\n'.format(k, v) for k, v in self._configuration().items()]) if self._configuration().items() else "\nParameters: Parameter-Free method\n") +
            "Attributes:\n\t" + ', '.join(c for c in (self.attributes if self.attributes is not None \
                else self.data.dataset_1.columns)) +
            "\nRuntime: {:2.4f} seconds".format(self.execution_time)
        )

    def get_weights_avg(self) -> float:
        return sum([w for _, _, w in self.pairs.edges(data='weight')])/len(self.pairs.edges(data='weight'))

    def get_weights_median(self) -> float:
        return [w for _, _, w in sorted(self.pairs.edges(data='weight'))][int(len(self.pairs.edges(data='weight'))/2)]    
    
    def get_weights_standard_deviation(self) -> float:
        return statistics.stdev([w for _, _, w in self.pairs.edges(data='weight')])
    
    def plot_distribution_of_all_weights(self) -> None:
        title = "Distribution of scores with " + self.metric + " metric in graph from entity matching"
        plt.figure(figsize=(10, 6))
        all_weights = [w for _, _, w in self.pairs.edges(data='weight')]
        sorted_weights = sorted(all_weights, reverse=True)
        
        plt.hist(sorted_weights)
        plt.xlim(0, 1)
        # only one line may be specified; full height
        plt.axvline(x = self.get_weights_avg(), color = 'blue', label = 'Average weight')
        plt.axvline(x = self.get_weights_median(), color = 'black', label = 'Median weight')
        plt.axvline(x = self.get_weights_avg()+self.get_weights_standard_deviation(), color = 'green', label = 'Average + SD weight')
        plt.legend()
        plt.show()

    def plot_distribution_of_all_weights_2d(self) -> None:
        title = "Distribution of scores with " + self.metric + " metric in graph from entity matching"
        plt.figure(figsize=(10, 6))
        all_weights = [w for _, _, w in self.pairs.edges(data='weight')]
        sorted_weights = sorted(all_weights, reverse=True)
        
        fig, ax = plt.subplots(tight_layout=True)
        hist = ax.hist2d(sorted_weights, sorted_weights)
        plt.axvline(x = self.get_weights_avg(), color = 'blue', label = 'Average weight')
        plt.axvline(x = self.get_weights_median(), color = 'black', label = 'Median weight')
        plt.axvline(x = self.get_weights_avg()+self.get_weights_standard_deviation(), color = 'green', label = 'Average + SD weight')
        plt.legend()
        plt.show()

    def plot_distribution_of_scores(self) -> None:
        title = "Distribution of scores with " + self.metric + " metric in graph from entity matching"
        def weight_distribution(G):
            bins = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
            distribution = [0] * (len(bins)-1)
            for u, v, w in G.edges(data='weight'):
                for i in range(len(bins) - 1):
                    if bins[i] <= w < bins[i + 1]:
                        distribution[i] += 1
                        break
            return distribution, len(G.edges(data='weight'))

        labels = [f'{(i)/10:.1f} - {(i+1)/10:.1f}' for i in range(0, 10)]

        distribution, num_of_pairs = weight_distribution(self.pairs)
        width = 0.5
        x = np.arange(len(labels))  # the label locations
        distribution = list(map(lambda x: (x/num_of_pairs)*100, distribution))
        print("Distribution-% of predicted scores: ", distribution)

        fig, ax = plt.subplots(figsize=(10,6))
        r1 = ax.bar(x, distribution, width, align='center', color='red')
        ax.set_xticks(x)
        ax.set_xticklabels(labels)
    
        
        # Add some text for labels, title and custom x-axis tick labels, etc.
        ax.set_ylabel('Percentage of pairs in each range to all (%)')
        ax.set_title(title)
        ax.set_xlabel('Similarity score range')
        fig.tight_layout()
        
        # only one line may be specified; full height
        plt.axvline(x = self.get_weights_avg()*10, color = 'blue', label = 'Average weight')
        plt.axvline(x = self.get_weights_median()*10, color = 'black', label = 'Median weight')
        plt.axvline(x = self.get_weights_avg()*10+self.get_weights_standard_deviation()*10, color = 'green', label = 'Average + SD weight')
        plt.legend()
        plt.show()

    def plot_gt_distribution_of_scores(self) -> None:
        title = "Distribution of scores with " + self.metric + " metric on ground truth pairs"
        def weight_distribution():
            bins = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
            distribution = [0] * (len(bins)-1)
            for _, (id1, id2) in self.data.ground_truth.iterrows():
                id1 = self.data._ids_mapping_1[id1]
                id2 = self.data._ids_mapping_1[id2] if self.data.is_dirty_er else self.data._ids_mapping_2[id2]
                w = self._calculate_vector_similarity(id1, id2)

                for i in range(len(bins) - 1):
                    if bins[i] <= w < bins[i + 1]:
                        distribution[i] += 1
                        break
            return distribution, len(self.data.ground_truth)

        labels = [f'{(i)/10:.1f} - {(i+1)/10:.1f}' for i in range(0, 10)]

        distribution, num_of_pairs = weight_distribution()
        width = 0.5
        x = np.arange(len(labels))  # the label locations
        distribution = list(map(lambda x: (x/num_of_pairs)*100, distribution))
        print("Distribution-% of predicted scores: ", distribution)

        fig, ax = plt.subplots(figsize=(10,6))
        r1 = ax.bar(x, distribution, width, align='center', color='blue')
        ax.set_xticks(x)
        ax.set_xticklabels(labels)

        # Add some text for labels, title and custom x-axis tick labels, etc.
        ax.set_ylabel('Percentage of pairs in each range to all (%)')
        ax.set_title(title)
        ax.set_xlabel('Similarity score range')
        fig.tight_layout()
        plt.show()

    def evaluate(self,
                 prediction,
                 export_to_df: bool = False,
                 export_to_dict: bool = False,
                 with_classification_report: bool = False,
                 verbose: bool = True) -> any:

        if self.data is None:
            raise AttributeError("Can not proceed to evaluation without data object.")

        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to evaluation without a ground-truth file. " +
                    "Data object has not been initialized with the ground-truth file")

        eval_obj = Evaluation(self.data)
        true_positives = 0
        total_matching_pairs = prediction.number_of_edges()
        for _, (id1, id2) in self.data.ground_truth.iterrows():
            id1 = self.data._ids_mapping_1[id1]
            id2 = self.data._ids_mapping_1[id2] if self.data.is_dirty_er \
                                                else self.data._ids_mapping_2[id2]
            if (id1 in prediction and id2 in prediction[id1]) or   \
                 (id2 in prediction and id1 in prediction[id2]):
                true_positives += 1

        eval_obj.calculate_scores(true_positives=true_positives,
                                  total_matching_pairs=total_matching_pairs)
        return eval_obj.report(self.method_configuration(),
                                export_to_df,
                                export_to_dict,
                                with_classification_report,
                                verbose)
        
    def stats(self) -> None:
        pass
    
    def export_pairs_to_csv(self, filename: str, with_similarity: bool = True) -> None:
        if self.pairs is None:
            raise AttributeError("Pairs have not been initialized yet. " +
                                 "Please run the method `run` first.")

        with open(filename, 'w') as f:
            for e1, e2, similarity in self.pairs.edges(data='weight'):
                e1 = self.data._ids_mapping_1[e1] if e1 < self.data.dataset_limit else self.data._ids_mapping_2[e1]
                e2 = self.data._ids_mapping_1[e2] if e2 < self.data.dataset_limit else self.data._ids_mapping_2[e2]
                if with_similarity:
                    f.write(f"{e1}, {e2}, {similarity}\n")
                else:
                    f.write(f"{e1}, {e2}\n")
            f.close()

    def export_to_df(self, prediction: Graph) -> pd.DataFrame:
        """creates a dataframe with the predicted pairs

        Args:
            prediction (any): Predicted graph

        Returns:
            pd.DataFrame: Dataframe with the predicted pairs
        """
        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to evaluation without a ground-truth file. \
                Data object mush have initialized with the ground-truth file")
        pairs_df = pd.DataFrame(columns=['id1', 'id2'])
        for edge in prediction.edges:
            id1 = self.data._gt_to_ids_reversed_1[edge[0]]
            id2 = self.data._gt_to_ids_reversed_1[edge[1]] if self.data.is_dirty_er \
                        else self.data._gt_to_ids_reversed_2[edge[1]]
            pairs_df = pd.concat([pairs_df, pd.DataFrame([{'id1':id1, 'id2':id2}], index=[0])], ignore_index=True)

        return pairs_df


class EntityMatching(AbstractEntityMatching):
    """Calculates similarity from 0.0 to 1.0 for all blocks
    """

    _method_name: str = "Entity Matching"
    _method_info: str = "Calculates similarity from 0. to 1. for all blocks"

    def __init__(
            self,
            metric: str = 'dice',
            tokenizer: str = 'white_space_tokenizer',
            vectorizer : str = None,
            qgram : int = 1,
            similarity_threshold: float = 0.5,
            tokenizer_return_unique_values = False, # unique values or not,
            attributes: any = None,
        ) -> None:
        super().__init__()
        self.pairs: Graph
        self.metric = metric
        self.attributes: list = attributes
        self.similarity_threshold = similarity_threshold
        self.tokenizer = tokenizer
        self.execution_time = 0
        self.vectorizer = vectorizer
        self.qgram: int = -1
        #
        # Selecting tokenizer
        #
        if metric not in available_metrics:
            raise AttributeError(
                'Metric ({}) does not exist. Please select one of the available. ({})'.format(
                    metric, available_metrics
                )
            )
        else:
            self._metric = metric

        self.tokenizer_return_set = (metric in set_metrics) or tokenizer_return_unique_values    
        self.qgram : int = qgram
        
        if(vectorizer is not None):
            if self.vectorizer not in available_vectorizers:
                raise AttributeError(
                    'Weighting Scheme ({}) does not exist. Please select one of the available. ({})'.format(
                        vectorizer, available_vectorizers
                    )
                )
        elif(tokenizer is not None):
            if tokenizer == 'white_space_tokenizer':
                self._tokenizer = WhitespaceTokenizer(return_set=self.tokenizer_return_set)
            elif tokenizer == 'char_tokenizer':
                self._tokenizer = QgramTokenizer(qval=self.qgram,
                                                return_set=self.tokenizer_return_set)
            elif tokenizer == 'word_tokenizer':
                self._tokenizer = WordQgramTokenizer(q=self.qgram)
            elif tokenizer not in available_tokenizers:
                raise AttributeError(
                    'Tokenizer ({}) does not exist. Please select one of the available. ({})'.format(
                        tokenizer, available_tokenizers
                    )
                )
        
    def predict(self,
                blocks: dict,
                data: Data,
                tqdm_disable: bool = False) -> Graph:
        """Main method of entity matching. Inputs a set of blocks and outputs a graph \
            that contains of the entity ids (nodes) and the similarity scores between them (edges).

            Args:
                blocks (dict): blocks of entities
                data (Data): dataset module
                tqdm_disable (bool, optional): Disables progress bar. Defaults to False.

            Returns:
                networkx.Graph: entity ids (nodes) and similarity scores between them (edges)
        """
        start_time = time()
        self.tqdm_disable = tqdm_disable
        
        if not blocks:
            raise ValueError("Empty blocks structure")
        self.data = data
        self.pairs = Graph()
        all_blocks = list(blocks.values())
        self._progress_bar = tqdm(total=len(blocks),
                                  desc=self._method_name+" ("+self.metric+ ", " + str(self.tokenizer) + ")",
                                  disable=self.tqdm_disable)
                
        if self.vectorizer is not None:
            self.initialize_vectorizer()

        if 'Block' in str(type(all_blocks[0])):
            self._predict_raw_blocks(blocks)
        elif isinstance(all_blocks[0], set):
            self._predict_candidate_pairs(blocks)
        else:
            raise AttributeError("Wrong type of Blocks")
        self.execution_time = time() - start_time
        self._progress_bar.close()

        return self.pairs
            
    def _predict_raw_blocks(self, blocks: dict) -> None:
        """Method for similarity evaluation blocks after Block building

        Args:
            blocks (dict): Block building blocks
        """
        if self.data.is_dirty_er:
            for _, block in blocks.items():
                entities_array = list(block.entities_D1)
                for index_1 in range(0, len(entities_array), 1):
                    for index_2 in range(index_1+1, len(entities_array), 1):
                        similarity = self._similarity(entities_array[index_1],
                                                      entities_array[index_2])
                        self._insert_to_graph(entities_array[index_1],
                                              entities_array[index_2],
                                              similarity)
                self._progress_bar.update(1)
        else:
            for _, block in blocks.items():
                for entity_id1 in block.entities_D1:
                    for entity_id2 in block.entities_D2:
                        similarity = self._similarity(entity_id1, entity_id2)
                        self._insert_to_graph(entity_id1, entity_id2, similarity)
                self._progress_bar.update(1)

    def initialize_vectorizer(self) -> None:
        self.frequency_evaluator : FrequencyEvaluator = FrequencyEvaluator(vectorizer=self.vectorizer,
                                                                            tokenizer=self.tokenizer,
                                                                            qgram=self.qgram)
        d1 = self.data.dataset_1[self.attributes] if self.attributes else self.data.dataset_1
        self._entities_d1 = d1 \
                    .apply(" ".join, axis=1) \
                    .apply(lambda x: x.lower()) \
                    .values.tolist()
        d2 = self.data.dataset_2[self.attributes] if self.attributes and not self.data.is_dirty_er else self.data.dataset_2
        self._entities_d2 = d2 \
                    .apply(" ".join, axis=1) \
                    .apply(lambda x: x.lower()) \
                    .values.tolist() if not self.data.is_dirty_er else None
        
        self.frequency_evaluator.fit(metric=self.metric,
                                    d1_entities=self._entities_d1,
                                    d2_entities=self._entities_d2)

    def _similarity(self, entity_id1: int, entity_id2: int) -> float:

        similarity: float = 0.0
        if self.vectorizer is not None:
            return self.frequency_evaluator.predict(id1=entity_id1, id2=entity_id2)
        elif isinstance(self.attributes, dict):
            for attribute, weight in self.attributes.items():
                e1 = self.data.entities.iloc[entity_id1][attribute].lower()
                e2 = self.data.entities.iloc[entity_id2][attribute].lower()

                similarity += weight*metrics_mapping[self._metric].get_sim_score(
                    self._tokenizer.tokenize(e1) if self._metric in set_metrics else e1,
                    self._tokenizer.tokenize(e2) if self._metric in set_metrics else e2
                )
        if isinstance(self.attributes, list):
            for attribute in self.attributes:
                e1 = self.data.entities.iloc[entity_id1][attribute].lower()
                e2 = self.data.entities.iloc[entity_id2][attribute].lower()
                similarity += metrics_mapping[self._metric].get_sim_score(
                    self._tokenizer.tokenize(e1) if self._metric in set_metrics else e1,
                    self._tokenizer.tokenize(e2) if self._metric in set_metrics else e2
                )
                similarity /= len(self.attributes)
        else:
            # concatenated row string
            e1 = self.data.entities.iloc[entity_id1].str.cat(sep=' ').lower()
            e2 = self.data.entities.iloc[entity_id2].str.cat(sep=' ').lower()
            te1 = self._tokenizer.tokenize(e1) if self._metric in set_metrics else e1
            te2 = self._tokenizer.tokenize(e2) if self._metric in set_metrics else e2
            similarity = metrics_mapping[self._metric].get_sim_score(te1, te2)
        return similarity

    def _configuration(self) -> dict:
        return  {
            "Metric" : self.metric,
            "Attributes" : self.attributes,
            "Similarity threshold" : self.similarity_threshold,
            "Tokenizer" : self.tokenizer,
            "Vectorizer" : self.vectorizer if self.vectorizer is not None else "None",
            "Qgrams" : self.qgram
        }

class VectorBasedMatching(AbstractEntityMatching):

    _method_name: str = "Vector Based Matching"
    _method_info: str = "Calculates similarity from 0. to 1. for vectors"

    def __init__(
            self,
            metric: str = 'cosine',
            similarity_threshold: float = 0.5,
        ) -> None:
        self.pairs: Graph
        self.metric = metric
        self.similarity_threshold = similarity_threshold
        self.vectors_d1 = None
        self.vectors_d2 = None
        self.execution_time = 0

        #
        # Selecting tokenizer
        #
        if metric not in vector_metrics:
            raise AttributeError(
                'Metric ({}) does not exist. Please select one of the available. ({})'.format(
                    metric, available_metrics
                )
            )
        else:
            self._metric = metric
            
    def predict(self,
                blocks: dict,
                data: Data,
                vectors_d1: np.array,
                vectors_d2: np.array = None,
                tqdm_disable: bool = False,
        ) -> Graph:
        """Main method of entity matching. Inputs a set of blocks and outputs a graph \
            that contains of the entity ids (nodes) and the similarity scores between them (edges).

            Args:
                blocks (dict): blocks of entities
                data (Data): dataset module
                tqdm_disable (bool, optional): Disables progress bar. Defaults to False.

            Returns:
                networkx.Graph: entity ids (nodes) and similarity scores between them (edges)
        """
        start_time = time()
        self.tqdm_disable = tqdm_disable
        self.vectors_d1 = vectors_d1
        # self.vectors_d2 = vectors_d2
        

        if(vectors_d1 is None):
            raise ValueError("Embeddings of the first dataset not given")
        else:
            self.vectors = vectors_d1
            if(not data.is_dirty_er):
                if(vectors_d2 is None):
                    raise ValueError("Embeddings of the second dataset not given")
                self.vectors = np.concatenate((vectors_d1,vectors_d2), axis=0)
        self.data = data
        self.pairs = Graph()
        self._progress_bar = tqdm(total=len(blocks),
                                  desc=self._method_name+" ("+self.metric+ ", " + str(self.tokenizer) + ")",
                                  disable=self.tqdm_disable)
        self._predict_candidate_pairs(blocks)
        self.execution_time = time() - start_time
        self._progress_bar.close()

        return self.pairs

    def _similarity(self, entity_id1: int, entity_id2: int) -> float:
        return vector_metrics_mapping[self._metric](self.vectors[entity_id1], self.vectors[entity_id2])

    def _configuration(self) -> dict:
        conf =  {
            "Metric" : self.metric,
            "Similarity threshold" : self.similarity_threshold
        }
        
        return conf
